;;; Wrappers that draw lines and boxes

(in-package :tutor)

(defmacro with-container ((container view) &body body)
  `(let ((,view (make-instance 'view :view-container ,container :view-position #@(0 0))))
     ,@body
     (resize-view-for-vertically ,view (coerce (view-subviews ,view) 'list))
     ,view))

(defun all-subviews (view &rest types)
  (let ((v (subviews view)))
    (when v
      (remove-if (complement (lambda (a) (loop for type in types
                                               when (typep a type)
                                               return t)))
                 (append v
                         (apply #'append (remove nil (mapcar (lambda (a) (apply 'all-subviews a types)) v))))))))

(defun line-height (font)
  "Return the line height for FONT."
  (let ((font-name (first font))
        (font-size (find-if 'numberp font)))
    (multiple-value-bind (ascent descent maxwidth leading)
        (font-info (list font-name font-size))
      (declare (ignore maxwidth))
      (+ ascent descent leading))))

(defun move-up (view amt)
  (let ((x (view-x view))
        (y (view-y view)))
    (set-view-position view x (- y amt))))

(defun abs-x (view)
  (point-h (convert-coordinates #@(0 0) view *window*)))
(defun abs-y (view)
  (point-v (convert-coordinates #@(0 0) view *window*)))

(defun hide-view (view)
  (set-view-position view (- (view-x view)) (- (view-y view))))

(defun unhide-view (view)
  (hide-view view))

(defun view-height (view)
  (point-v (view-size view)))

(defun view-width (view)
  (point-h (view-size view)))

(defun view-x (view)
  (point-h (view-position view)))

(defun view-y (view)
  (point-v (view-position view)))

(defun view-bottom (view)
  (+ (view-y view) (view-height view)))

(defun view-right (view)
  (+ (view-x view) (view-width view)))

(defun view-center (view)
  "Return the center of the view relative to its 0,0 point."
  (make-point (truncate (view-width view) 2)
              (truncate (view-height view) 2)))

(defun rect-center (x y w h)
  "Return the center of the view given the corners of the rectangle."
  (make-point (+ x (truncate w 2))
              (+ y (truncate h 2))))

(defun center-view-in-view (v1 v2 &optional (x (view-x v1)))
  "Center v1 inside v2 vertically."
  (set-view-position v1 x (- (truncate (view-height v2) 2)
                             (truncate (view-height v1) 2))))

(defun find-alignment (view)
  (let ((align (loop for v in (coerce (view-subviews view) 'list)
		     when (typep v 'alignment-view)
		     return v)))
    (if align
	(view-x align)
      0)))

(defun order-views (views &optional (pad 0) (y 0) (offset 0))
  (let ((max (apply #'max (mapcar 'view-height views))))
    (loop for v in views
          with i = offset
          do (set-view-position v i (+ y (- (truncate max 2) (truncate (view-height v) 2))))
          do (incf i (+ (view-width v) pad)))))


(defun resize-view-for (view views)
  (set-view-size view
                 (apply #'+ (mapcar 'view-width views))
                 (apply #'max (mapcar 'view-height views))))

(defun order-views-vertically (views)
  (loop for v in views
	with i = 0
	with align = (find-alignment (first views))
	do (set-view-position v (- align (find-alignment v)) i)
	do (incf i (view-height v))))

(defun resize-view-for-vertically (view views)
  (let ((x (apply #'min (mapcar 'view-x views)))
	(y (apply #'min (mapcar 'view-y views)))
	(mx (apply #'max (mapcar (lambda (v)
				   (+ (view-width v)
				      (view-x v)))
				 views)))
	(my (apply #'max (mapcar (lambda (v)
				   (+ (view-height v)
				      (view-y v)))
				 views))))
    (set-view-size view (- mx x) (- my y))
    (set-view-scroll-position view x y)))

(defclass geom (simple-view)
  ((color :initarg :color :accessor color)))

(defclass line (geom)
  ((x1 :initarg :x1 :accessor x1)
   (x2 :initarg :x2 :accessor x2)
   (y1 :initarg :y1 :accessor y1)
   (y2 :initarg :y2 :accessor y2)))

(defclass hilite-line (line)
  ((enter-fn :initarg :enter-fn :accessor enter-fn)
   (leave-fn :initarg :leave-fn :accessor leave-fn)
   (hilite-color :initarg :hilite-color :accessor hilite-color)
   (id :initarg :id :accessor hilite-id)))

(defclass pointer (geom)
  ((x1 :initarg :x1 :accessor x1)
   (x2 :initarg :x2 :accessor x2)
   (y1 :initarg :y1 :accessor y1)
   (y2 :initarg :y2 :accessor y2)))

(defclass rect (geom)
  ((x1 :initarg :x1 :accessor x1)
   (x2 :initarg :x2 :accessor x2)
   (y1 :initarg :y1 :accessor y1)
   (y2 :initarg :y2 :accessor y2)
   (solid-p :initarg :solid-p :accessor solid-p)))

(defclass oval (geom)
  ((x1 :initarg :x1 :accessor x1)
   (x2 :initarg :x2 :accessor x2)
   (y1 :initarg :y1 :accessor y1)
   (y2 :initarg :y2 :accessor y2)))

(defclass arrow (geom) 
  ((fill-p :initform nil :initarg :fill-p :accessor arrow-fill-p)
   (polygon :accessor arrow-polygon)))

(defclass text (geom)
  ((txt :initform nil :initarg :txt :accessor text-txt)
   (ascent :initform nil :initarg :ascent :accessor text-ascent)))

(defclass user-instruction (text)
  ((initial-text :initarg :initial-text :accessor initial-text)
   (user-text :initarg :user-text :accessor user-text)))

(defclass user-hint-instruction (user-instruction)
  ())

(defmethod view-draw-contents ((line line))
  (let ((cont (view-container line)))
    (with-fore-color (color line)
      (move-to cont (make-point (x1 line) (y1 line)))
      (line-to cont (make-point (x2 line) (y2 line))))))

(defmethod view-draw-contents ((line pointer))
  (let* ((cont (view-container line))
         (h (- (y2 line) (y1 line)))
         (w (- (x2 line) (x1 line)))
         (angle (if (= w 0)
                    (/ pi 2)
                  (atan (/ h w))))
         (angle1 (- angle (/ pi 8)))
         (angle2 (+ angle (/ pi 8)))
         (sign (if (< w 0) -1 1))
         (a1x (round (- (x2 line) (* sign 10 (cos angle1)))))
         (a1y (round (- (y2 line) (* sign 10 (sin angle1)))))
         (a2x (round (- (x2 line) (* sign 10 (cos angle2)))))
         (a2y (round (- (y2 line) (* sign 10 (sin angle2)))))
         poly)
    (with-fore-color (color line)
      (move-to cont (x1 line) (y1 line))
      (line-to cont (x2 line) (y2 line))
      (start-polygon cont)
      (line-to cont a1x a1y)
      (line-to cont a2x a2y)
      (line-to cont (x2 line) (y2 line))
      (setf poly (get-polygon cont))
      (paint-polygon cont poly)
      (event-dispatch)
      (kill-polygon poly))))

(defmacro with-pen-size (view h v &body body)
  (with-gensyms (tmp)
    `(let ((,tmp (pen-size ,view)))
       (unwind-protect
           (progn
             (set-pen-size ,view ,h ,v)
             ,@body)
         (set-pen-size ,view ,tmp)))))

(defmethod view-draw-contents ((rect rect))
  (let ((cont (view-container rect)))
    (with-fore-color (color rect)
      (if (solid-p rect)
          ;; instead of filling it completely. just draw it a few pixels thick
          (with-pen-size cont 3 3
            (paint-rect cont (x1 rect) (y1 rect) (x2 rect) (y2 rect)))
        (frame-rect cont (x1 rect) (y1 rect) (x2 rect) (y2 rect))))))


(defmethod view-draw-contents ((oval oval))
  (let ((cont (view-container oval)))
    (with-fore-color (color oval)
      (frame-oval cont (x1 oval) (y1 oval) (x2 oval) (y2 oval)))))

(defmethod view-draw-contents ((arrow arrow))
  (with-fore-color (color arrow)
    (if (arrow-fill-p arrow)
        (paint-polygon arrow (arrow-polygon arrow))
      (frame-polygon arrow (arrow-polygon arrow)))))

(defmethod view-draw-contents ((text text))
  (with-fore-color (color text)
    (move-to text 0 (text-ascent text))
    (princ (text-txt text) text)))

(defmethod view-draw-contents ((view user-instruction))
  (with-fore-color *lightest-gray-color*
    (paint-rect view 0 0 (view-width view) (view-height view)))
  (with-fore-color (color view)
    (frame-rect view 1 1 (1- (view-width view)) (1- (view-height view)))
    (move-to view 5 (text-ascent view))
    (princ (wrap-text (or (user-text view)
                          (initial-text view))
                      (view-width view)
                      (view-font view))
           view)))

;; All this hilite stuff is a giant hack. be warned.
(defun draw-hilite-line (view x1 y1 x2 y2 id enter-fn leave-fn &optional (color *foreground-color*))
  (let* ((pos (make-point (min x1 x2) (min y1 y2)))
	 (size (subtract-points (add-points #@(1 1) (make-point (max x1 x2) (max y1 y2))) pos)))
    (make-instance 'hilite-line
		   :x1 x1
		   :y1 y1
		   :x2 x2
		   :y2 y2
		   :color color
		   :view-container view
		   :view-position pos
		   :view-size size
                   :id id
                   :hilite-color color
                   :enter-fn enter-fn
                   :leave-fn leave-fn)))

(defun draw-line (view x1 y1 x2 y2 &optional (color *foreground-color*))
  (let* ((pos (make-point (min x1 x2) (min y1 y2)))
	 (size (subtract-points (add-points #@(1 1) (make-point (max x1 x2) (max y1 y2))) pos)))
    (make-instance 'line
		   :x1 x1
		   :y1 y1
		   :x2 x2
		   :y2 y2
		   :color color
		   :view-container view
		   :view-position pos
		   :view-size size)))

(defun draw-pointer (view x1 y1 x2 y2 &optional (color *foreground-color*))
  (let* ((pos (make-point (min x1 x2) (min y1 y2)))
	 (size (subtract-points (add-points #@(1 1) (make-point (max x1 x2) (max y1 y2))) pos)))
    (make-instance 'pointer
		   :x1 x1
		   :y1 y1
		   :x2 x2
		   :y2 y2
		   :color color
		   :view-container view
                   ;; Grow the bounding box by 10 so the arrow heads get drawn
		   :view-position (subtract-points pos #@(5 5))
		   :view-size (add-points size #@(10 10)))))

(defun draw-rect (view x1 y1 x2 y2 &optional (color *foreground-color*) solid-p)
  (let* ((pos (make-point (min x1 x2) (min y1 y2)))
	 (size (subtract-points (make-point (max x1 x2) (max y1 y2)) pos)))
    (make-instance 'rect
		   :x1 x1
		   :y1 y1
		   :x2 x2
		   :y2 y2
		   :color color
                   :solid-p solid-p
		   :view-container view
		   :view-position pos
		   :view-size size)))

(defun draw-oval (view x1 y1 x2 y2 &optional (color *foreground-color*))
  (let* ((pos (make-point (min x1 x2) (min y1 y2)))
	 (size (subtract-points (make-point (max x1 x2) (max y1 y2)) pos)))
    (make-instance 'oval
		   :x1 x1
		   :y1 y1
		   :x2 x2
		   :y2 y2
		   :color color
		   :view-container view
		   :view-position pos
		   :view-size size)))

(defun draw-text-plus-header (view x y header txt font &optional (hcolor *red-color*) (color *foreground-color*))
  (let* ((parent (make-instance 'view
                                :view-position (make-point x y)
                                :view-container view))
         (h (make-instance 'static-text-dialog-item
                           :part-color-list `(:text ,hcolor)
                           :view-container parent
                           :view-font font
                           :view-position #@(0 0)
                           :dialog-item-text header))
         (b (make-instance 'static-text-dialog-item
                           :part-color-list `(:text ,color)
                           :view-container parent
                           :view-font font
                           :view-position (make-point 0 (point-v (view-size h)))
                           :dialog-item-text txt)))
    (set-view-size parent (max (point-h (view-size h))
                               (point-h (view-size b)))
                   (+ (point-v (view-size h))
                      (point-v (view-size b))))
    parent))

(defun draw-text (view x y txt &optional (color *foreground-color*) part-list-p)
  (make-instance 'static-text-dialog-item
		 :part-color-list (if part-list-p
                                        color
                                      `(:text ,color))
                 :view-font *default-font*
		 :view-container view
		 :view-position (make-point x y)
		 :dialog-item-text txt))

(defun draw-transparent-text (view x y txt &optional (font *default-font*) (color *foreground-color*))
  (multiple-value-bind (ascent descent maxwidth leading) (font-info font)
    (declare (ignore descent maxwidth leading))
    (make-instance 'text
                   :view-container view
                   :txt txt
                   :ascent ascent
                   :view-font font
                   :color color
                   :view-position (make-point x y)
                   :view-size (make-point (string-width txt font) (line-height font)))))

(defun draw-button (view x y txt action)
  (make-instance 'button-dialog-item
                 :view-container view
                 :view-position (make-point x y)
                 :dialog-item-text txt
                 :dialog-item-action action))

(defun draw-text-with-font (view x y txt font &optional (color *foreground-color*) part-list-p)
  (make-instance 'static-text-dialog-item
		 :part-color-list (if part-list-p
                                        color
                                      `(:text ,color))
		 :view-font font
		 :view-container view
		 :view-position (make-point x y)
		 :dialog-item-text txt))

(defun draw-selected-text-with-font (view x y txt font &optional (color *foreground-color*))
  (let* ((v (make-instance 'view
                           :view-container view)))
    (draw-oval v x y (+ x 10) (+ y 10))
    (make-instance 'static-text-dialog-item
                   :part-color-list `(:text ,color)
                   :view-font font
                   :view-container v
                   :view-position (make-point x y)
                   :dialog-item-text txt)
    v))

(defclass hilite-text (static-text-dialog-item)
  ((enter-fn :initarg :enter-fn :accessor enter-fn)
   (leave-fn :initarg :leave-fn :accessor leave-fn)
   (hilite-color :initarg :hilite-color :accessor hilite-color)
   (id :initarg :id :accessor hilite-id)))

(defmethod view-mouse-enter-event-handler ((view hilite-text))
  (when (enter-fn view)
    (funcall (enter-fn view) view)))

(defmethod view-mouse-leave-event-handler ((view hilite-text))
  (when (leave-fn view)
    (funcall (leave-fn view) view)))

(defun draw-hilite-text (view x y txt font id enter-fn leave-fn &optional (color *foreground-color*) part-list-p)
  (let ((v (make-instance 'hilite-text
                          :hilite-color (if part-list-p
                                            (copy-list color)
                                          `(:text ,color))
                          :part-color-list (if part-list-p
                                               color
                                             `(:text ,color))
                          :view-font font
                          :view-container view
                          :view-position (make-point x y)
                          :dialog-item-text txt
                          :id id
                          :enter-fn enter-fn
                          :leave-fn leave-fn)))
    v))

;; (defun draw-arrow (x y &optional (color *foreground-color*))
;;   (draw-line
(defun draw-fraction (n x y container id enter-fn leave-fn &optional (color *foreground-color*) part-list-p)
  (let* ((v (make-instance 'view
                           :view-container container))
         (num (draw-hilite-text v x y (format nil "~a" (abs (numerator n))) +eqn-font+ id enter-fn leave-fn color part-list-p))
         (den (draw-hilite-text v x y (format nil "~a" (denominator n)) +eqn-font+ id enter-fn leave-fn color part-list-p))
         (minus (when (< n 0)
                  (draw-hilite-text v x y "-" +eqn-font+ id enter-fn leave-fn color part-list-p)))
         (offset (if minus (view-width minus) 0))
         (max (max (view-width num) (view-width den))))
    (set-view-size v (+ max offset)
                   (+ (view-height num) (view-height den)))
    (when minus
      (center-view-in-view minus v 0))
    (set-view-position num (+ offset (- (truncate max 2) (truncate (view-width num) 2))) 0)
    ;; FIXME: the color thing is a TOTAL hack. if part-list-p is true
    ;; then its selected and if its selected it should be red.
    (draw-hilite-line v offset (view-height num) (+ offset max) (view-height num) id enter-fn leave-fn (if part-list-p (getf color :text) color))
    (set-view-position den (+ offset (- (truncate max 2) (truncate (view-width den) 2))) (+ (view-height num) 1))
    v))

(defun draw-number (n container x y id enter-fn leave-fn &optional (color *foreground-color*) part-list-p)
  (if (typep n 'ratio)
      (draw-fraction n x y container id enter-fn leave-fn color part-list-p)
    (draw-hilite-text container x y (prin1-to-string n) +eqn-font+ id enter-fn leave-fn color part-list-p)
;;     (make-instance 'draw-hilite-text
;;                    :part-color-list (if part-list-p
;;                                         color
;;                                       `(:text ,color))
;;                    :view-font +eqn-font+
;;                    :view-container container
;;                    :dialog-item-text (prin1-to-string n))
    ))

(defun delete-subviews (view)
  (apply #'remove-subviews view (coerce (view-subviews view) 'list)))

(defun delete-subviews-if (view fn)
  (dolist (v (coerce (view-subviews view) 'list))
    (when (funcall fn v)
      (remove-subviews view v))))

(defmacro create-widget-set (name x y &key (fmt "~A") (view *window*) font sexpr)
  (let ((var (intern (format nil "*~a-WIDGET*" (symbol-name name))))
	(init (intern (format nil "INIT-~a-WIDGET" (symbol-name name))))
	(clear (intern (format nil "CLEAR-~a" (symbol-name name))))
	(draw (intern (format nil "DRAW-~a" (symbol-name name))))
	(txt (gensym)))
    `(progn
       (defvar ,var nil)
       (defun ,init ()
	 (setf ,var nil))
       (defun ,clear ()
	 (when ,var
	   (remove-subviews ,view ,var)
	   (,init)))
       (defun ,draw (&rest ,txt)
	 (,clear)
	 (setf ,var (draw-text-with-font ,view ,x ,y (apply #'format nil ,fmt ,txt) ,font))
         ,sexpr))))

;; (create-widget-set demerits (+ +side-bar-x+ 5) 370 :fmt "Demerits: ~A" :view *window*)
(defun init-demerits-widget ())
(defun clear-demerits ())
(defun draw-demerits (&rest txt) (declare (ignore txt)))

;;(create-widget-set message 0 0 :view *window* :font '("Monaco" :bold 12) :sexpr )

(create-widget-set problems-left 0 0 :view *window*)

(defun draw-arrow (view x y length height direction &optional (color *foreground-color*) 
		   &aux (niph (truncate height 5))
		        (h (- height niph niph))
		        (tipw (truncate length 2))
			(rodw (- length tipw))
			(st1 (make-point 0 niph))
 		        (rod1 (make-point rodw niph))
		        (nip1 (make-point (point-h rod1) 0))
			(tip (make-point length (truncate height 2)))
			(nip2 (add-points nip1 (make-point 0 height)))
			(rod2 (add-points rod1 (make-point 0 h)))
			(st2 (make-point 0 (- height niph))))
  (let ((lv (make-instance 'arrow
                           :fill-p t
                           :color color
			   :view-position (make-point x y)
			   :view-size (make-point length (1+ height))
			   :view-container view)))
    (labels ((trans (p)
                    (if (eq direction :left)
                        (make-point (+ length (* (point-h p) -1)) (point-v p))
                      p)))
      (start-polygon lv)
      (move-to lv (trans st1))
      (line-to lv (trans rod1))
      (line-to lv (trans nip1))
      (line-to lv (trans tip))
      (line-to lv (trans nip2))
      (line-to lv (trans rod2))
      (line-to lv (trans st2))
      (line-to lv (trans st1))
      (setf (arrow-polygon lv) (get-polygon lv))
      lv)))

(defun draw-vert-arrow (view x y length height &optional (color *foreground-color*) 
                             &aux (nipw (truncate length 5))
                             (w (- length nipw nipw))
                             (tiph (truncate height 2))
                             (rodh (- height tiph))
                             (st1 (make-point nipw height))
                             (rod1 (make-point nipw (- height rodh)))
                             (nip1 (make-point 0 (point-v rod1)))
                             (tip (make-point (truncate length 2) 0))
                             (nip2 (add-points nip1 (make-point length 0)))
                             (rod2 (add-points rod1 (make-point w 0)))
                             (st2 (make-point (- length nipw) height)))
  "X,Y is the corner of the ??. it points up:
   /\
   ||"
  (let ((lv (make-instance 'arrow
                           :fill-p t
                           :color color
			   :view-position (make-point x y)
			   :view-size (make-point length (1+ height))
			   :view-container view)))
;;     (labels ((dl (p1 p2)
;; 	       (draw-line lv (point-h p1) (point-v p1) (point-h p2) (point-v p2) color)))
    (start-polygon lv)
    (move-to lv st1)
    (line-to lv rod1)
    (line-to lv nip1)
    (line-to lv tip)
    (line-to lv nip2)
    (line-to lv rod2)
    (line-to lv st2)
    (line-to lv st1)
    (setf (arrow-polygon lv) (get-polygon lv))
    lv))

;; 
(defclass box-view (view) ())

;; prototype = BUTTON | HBOX | VBOX | TEXT | VIEW | SPACE
;; BUTTON = (:button BTEXT ACTION :name SYMBOL)
;; HBOX = (:hbox CONTENTS PAD :width NUMBER :fill BOOL :border BOOL :name SYMBOL)
;; VBOX = (:vbox CONTENTS PAD :height NUMBER :fill BOOL :border BOOL :name SYMBOL)
;; TEXT = (:text STRING :name SYMBOL)
;; VIEW = (:view :width NUMBER :height NUMBER :border BOOL :name SYMBOL)
;; SPACE = (:space :width NUMBER :height NUMBER)
;; ATTR = 
;; PAD = :pad NUMBER
;; CONTENTS = (PROTOTYPE+)

(defun create-layout (prototype container &key (x 0)
                                               (y 0)
                                               (w (point-h (view-size container)))
                                               (h (point-v (view-size container))))
    (labels ((handle-button (props)
             (let ((widget (make-instance 'button-dialog-item
                                          :view-container container
                                          :view-position (make-point (+ x 4) (1+ y))
                                          :view-size (make-point (- w 8) (- h 8))
                                          :dialog-item-enabled-p (not (getf props :disabled))
                                          :dialog-item-text (getf props :text)
                                          :dialog-item-action (getf props :action))))
               (when (getf props :name)
                 (list (cons (getf props :name) widget)))))
             (handle-text (props)
               (let ((widget (apply 'make-instance 'static-text-dialog-item
                                    :part-color-list `(:text ,*foreground-color*)
                                    :view-container container
                                    :view-position (make-point x y)
                                    :view-size (make-point w h)
                                    :dialog-item-text (getf props :text)
                                    (when (getf props :font)
                                      (list :view-font (getf props :font))))))
                 (when (getf props :name)
                   (list (cons (getf props :name) widget)))))
             (handle-view (props)
               ;; The border must be drawn first otherwise it steals events
               (when (getf props :border)
                 (draw-rect container x y (+ x w) (+ y h))
                 (incf x) (incf y) (decf w 2) (decf h 2))
               (let ((widget (make-instance 'view
                                            :view-container container
                                            :view-position (make-point x y)
                                            :view-size (make-point w h))))
                 (when (getf props :name)
                   (list (cons (getf props :name) widget)))))
             (handle-hbox (props)
               ;; The border must be drawn first otherwise it steals events
               (when (getf props :border)
                 (draw-rect container x y (+ x w) (+ y h))
                 (incf x) (incf y) (decf w 2) (decf h 2))
               (let* ((widget (make-instance 'view
                                              :view-container container
                                              :view-position (make-point x y)
                                              :view-size (make-point w h)))
                      (contents (getf props :contents))
                      (pad (or (getf props :pad) 0))
                      (hard (remove-if (lambda (x)
                                         (not (getf (cdr x) :width))) contents))
                      (soft (remove-if (lambda (x)
                                         (getf (cdr x) :width)) contents))
                      ;; space left over to divide among soft
                      (space (- w (or (loop for i in hard summing (getf (cdr i) :width)) 0)
                                (* pad (1- (length contents)))))
                      (space-per (truncate space (length soft)))
                      (xofs 0))
                 (delete-if #'null
                            (nconc (when (getf props :name)
                                     (list (cons (getf props :name) widget)))
                                   (apply #'nconc (loop for c in contents
                                                         collect (if (find c soft :test #'equal)
                                                                     (prog1 (create-layout c widget :x xofs :y 0 :w space-per :h h)
                                                                       (incf xofs (+ space-per pad)))
                                                                   (prog1
                                                                       (create-layout c widget :x xofs :y 0 :w (getf (cdr c) :width) :h h)
                                                                     (incf xofs (+ (getf (cdr c) :width) pad))))))))))
             (handle-vbox (props)
               ;; The border must be drawn first otherwise it steals events
               (when (getf props :border)
                 (draw-rect container x y (+ x w) (+ y h))
                 (incf x) (incf y) (decf w 2) (decf h 2))
               (let* ((widget (make-instance 'view
                                              :view-container container
                                              :view-position (make-point x y)
                                              :view-size (make-point w h)))
                      (contents (getf props :contents))
                      (pad (or (getf props :pad) 0))
                      (hard (remove-if (lambda (x)
                                         (not (getf (cdr x) :height))) contents))
                      (soft (remove-if (lambda (x)
                                         (getf (cdr x) :height)) contents))
                      ;; space left over to divide among soft
                      (space (- h (or (loop for i in hard summing (getf (cdr i) :height)) 0)
                                (* pad (1- (length contents)))))
                      (space-per (truncate space (length soft)))
                      (yofs 0))
                 (delete-if #'null
                            (nconc (when (getf props :name)
                                     (list (cons (getf props :name) widget)))
                                   (apply #'nconc (loop for c in contents
                                                         collect (if (find c soft :test #'equal)
                                                                     (prog1 (create-layout c widget :x 0 :y yofs :w w :h space-per)
                                                                       (incf yofs (+ space-per pad)))
                                                                   (prog1
                                                                       (create-layout c widget :x 0 :y yofs :w w :h (getf (cdr c) :height))
                                                                     (incf yofs (+ (getf (cdr c) :height) pad)))))))))))
      (ecase (first prototype)
        (:space nil)
        (:button
         (handle-button (rest prototype)))
        (:hbox
         (handle-hbox (rest prototype)))
        (:vbox
         (handle-vbox (rest prototype)))
        (:text
         (handle-text (rest prototype)))
        (:view
         (handle-view (rest prototype))))))


(defun blink-view (view &optional (duration 100))
  (invert-rect view 0 0 (view-width view) (view-height view))
  (tutor-sleep duration)  
  (invert-rect view 0 0 (view-width view) (view-height view)))

(defun draw-user-instruction (view x y width height &optional initial-text user-text hint-p resize-p (font *default-font*) (color *foreground-color*))
  (multiple-value-bind (ascent descent maxwidth leading) (font-info font)
    (declare (ignore descent maxwidth leading))
    (make-instance (if hint-p 
                       'user-hint-instruction
                       'user-instruction)
                   :view-container view
                   :view-position (make-point x y)
                   :view-size (make-point width (if resize-p (text-height user-text font width) height))
                   :initial-text initial-text
                   :user-text user-text
                   :ascent ascent
                   :view-font font
                   :color color)))

(defun text-height (text font width)
  (* (or (and text
              (1+ (count #\Newline (wrap-text text width font))))
         1)
     (line-height font)))

(defun resize-user-instruction-for-text (view)
  (set-view-size view (view-width view) (text-height (user-text view) (view-font view) (view-width view)))
  view)
