;;; History support

(in-package :tutor)

;; (defvar *history-id* nil)

(defmacro with-history-file ((s) &body body)
  `(without-interrupts
       (with-open-file (,s (merge-pathnames (make-pathname  :name (format nil "history-~a" *history-id*) :type "txt")
                                            *history-path*)
                           :direction :output
                           :if-exists :append
                           :if-does-not-exist :create)
         ,@body)))

(defun history-init (id)
  (setf *history-id* id)
  (with-history-file (stream)
    (multiple-value-bind (s min h d m y) (decode-universal-time (get-universal-time))
      (format stream ";; Tutor started at ~a:~a:~a ~a/~a/~a~%" h min s d m y))))

(defun history-end ()
  (with-history-file (stream)
    (multiple-value-bind (s min h d m y) (decode-universal-time (get-universal-time))
      (format stream ";; Experiment ended at ~a:~a:~a ~a/~a/~a (score ~a)~%" h min s d m y (tutor-score *tutor*)))))

(defun history-add (type &rest args)
  ;; get-time is defined if act-r is loaded. The assumption is if you
  ;; *have act-r loaded then you're running the act-r component.
  (let ((time (if (fboundp 'common-lisp-user::get-time)
                  ;; use funcall to get rid of the warning when it isn't defined
                  (funcall 'common-lisp-user::get-time)
                (get-internal-real-time)))
        ;; This is a hack to make sure TUTOR:: doesn't get prefixed on
        ;; all the symbols.
        (*package* (find-package :tutor)))
  (with-history-file (s)
    (format s "~s~%" `(,type ,time ,@args)))))

;; (defmacro history-type (name type &rest args)
;;   `(defun ,(intern (format "HISTORY-~a" (symbol-name name))) ,args
;;      (history-add ,type ,@args)))

(defun history-add-start-tutor (extra)
  (apply 'history-add 'start-tutor
         :date (current-date)
         :version *version*
         :id *history-id*
         :customs (get-customizable-values)
         extra))

(defun history-add-operation (name clicked tree new-tree)
  "Add a click event to the history."
  (history-add 'operation name 'successful (if clicked 'clicked 'typed) tree new-tree))
  
(defun history-add-bad-operation (name clicked)
  "Add an operation event that was unsuccessful to the history."
  (history-add 'operation name 'unsuccessful (if clicked 'clicked 'typed) (box-to-tree (tutor-eqn-root *tutor*) t t)))

(defun history-add-submit ()
  "the 3rd argument is the tree after the submit."
  (history-add 'submit (tutor-incorrect-submit *tutor*) (box-to-tree (tutor-eqn-root *tutor*) nil t)))

(defun history-add-bad-submit ()
  "Logged when the user typed answers, auto-whats-next was on,
  and they typed something wrong. The 3rd argument is the tree
  after the submit."
  (history-add 'bad-submit))

(defun history-add-whats-next (txt)
  "the 3rd argument is the tree after the submit."
  (history-add 'whats-next txt))

(defun history-go-back (clicked state)
  "The user decided to go back."
  (history-add 'back (if clicked 'clicked 'typed) state))

(defun history-go-forward (clicked state)
  (history-add 'forward (if clicked 'clicked 'typed) state))

(defun history-first-mistake (mistakes-p)
  "The user decided to go back to the last mistake. MISTAKES-P is
  T when there were mistakes."
  (history-add 'first-mistake mistakes-p))

(defun history-first-mistake-fail ()
  "logged when the user tried to fist the last mistake while in
  submit mode."
  (history-add 'first-mistake-fail))

(defun history-add-input (arg val raw)
  (let* ((box (arg-container arg (tutor-eqn-root *tutor*)))
	 (dump (dump-box box)))
    ;; Specify which arg is being filled in
    (when (typep box 'arg-box)
      (cond
       ((eq arg (box-arg1 box))
	 (setf (first (second dump)) 'fill))
	((eq arg (box-arg2 box))
	 (setf (third (second dump)) 'fill))
	((eq arg (box-op box))
	 (setf (second (second dump)) 'fill))))
    (history-add 'input val raw dump)))

(defun history-add-box-input (box val raw)
  (history-add 'box-input val raw (dump-box box)))

(defun history-log-input (ch)
  (history-add 'typed-input ch))

(defun history-add-keypress (key)
  (history-add 'keypress key))

(defun history-add-click (match x y thing)
  "add a click event. match is true when the click actually did
something. The next history item will be what they did."
  (apply 'history-add 'click match x y thing))

(defun history-select (box)
  (history-add 'select (dump-box box)))

(defun history-unselect (box)
  (history-add 'unselect (dump-box box)))

(defun history-add-next-prob (correctp clicked)
  (if correctp
      (history-add 'problem-end (tutor-score *tutor*) (if clicked 'clicked 'typed))
    (history-add 'incorrect-answer (tutor-score *tutor*) (if clicked 'clicked 'typed))))

(defun history-add-new-prob (tree)
  (history-add 'problem-start tree))

;; (defun pprint-history (history)
;;   (loop for i in history
;; 	do (format t "~{~16,,s|~}~%" i)))

;; history for the calculator

(defun history-add-calc-press (value clickp)
  (history-add 'calc value clickp))

;; For the text portion

(defun history-add-next ()
  (history-add 'text-next))

(defun history-add-prev ()
  (history-add 'text-previous))

(defun history-add-open-dialog (type id)
  (history-add 'open-dialog type id))

(defun history-add-close-dialog ()
  (history-add 'close-dialog))

(defun history-add-break ()
  (history-add 'break))

(defun history-add-opn-during-submit (op clicked)
  (history-add 'opn-during-submit op (if clicked 'clicked 'typed)))

(defun history-add-failed-nav (clicked forward)
  (if forward
      (history-add 'bad-forward (if clicked 'clicked 'typed))
    (history-add 'bad-back (if clicked 'clicked 'typed))))

;; other

(defun history-show-instructions ()
  (history-add 'show-instructions))
(defun history-end-instructions ()
  (history-add 'end-instructions))

;; 0back

(defun history-0back-click ()
  (history-add '0back-click))

(defun history-0back-letter (letter)
  (history-add '0back-letter letter))

(defun history-0back-crosshair ()
  (history-add '0back-crosshair))

(defun history-0back-start ()
  (history-add '0back-start))

(defun history-0back-end ()
  (history-add '0back-end))

(defun history-0back-wait ()
  (history-add '0back-wait))

;;; mouse input

(defun history-input-open ()
  (history-add 'input-open))

(defun history-input-close ()
  (history-add 'input-close))

(defun history-input-click-button (button)
  (history-add 'input-click button))

(defun history-enter-expression (id)
  (history-add 'mouse-enter id))

(defun history-leave-expression (id)
  (history-add 'mouse-leave id))
