(declaim (optimize (debug 3) (speed 0) (size 0)))

(in-package :tutor)

;; (defvar *input-dialog* nil
;;   "The dialog box for typing input.")

(defun clear-input-dialog ()
  "Call this if the fill-in stage must be aborted. This happens if the user presses <-."
  (when *input-window*
    (if (typep *input-window* 'input-window)
        (remove-subviews (view-container *input-window*) *input-window*)
        (window-close *input-window*))
    (handle-window-close *input-window*)))

;; FIXME: the user-supplied-hint stuff breaks this input type.
(defun read-dialog-input (x y title text post-fn thing &optional (window *window*))
  (declare (ignore x y window thing))
  (let (dialog)
    ;; If the user clicked a green box while the dialog was still
    ;; open, close it then proceed. Otherwise they may get confused
    ;; because they click the green box and nothing happens. Little do
    ;; they know a dialog window is hiding behind the tutor window.
    (when *input-window*
      (window-close *input-window*)
      (setf *input-window* nil))
    (labels ((on-submit (input)
                        (window-close dialog)
                        (funcall post-fn input)
                        (handle-window-close dialog)))
      (setf dialog (get-string-from-user text
                                         :position (make-point (truncate (tutor-width) 2) (tutor-height))
                                         :window-title title
                                         :size #@(400 100)
                                         :modeless t
                                         :action-function #'on-submit))
      (setf *input-window* dialog)
      ;; for finegrained hints
      (update-hints))))

;;;; pop up a window and ask for a number or an operator


;; EQN = NUMBER OP NUMBER
;; 

(defmacro expect (test &body body)
  `(if ,test
       (progn ,@body)
     (error "Parse failure: ~a" ,test)))

(defun parse-number (stream)
  (let ((acc (make-array 1 :element-type 'character :adjustable t :fill-pointer 0)))
    ;; Read a possible -
    (when (char= #\- (peek-char nil stream))
      (vector-push-extend (read-char stream) acc))
    ;; Read the number
    (expect (digit-char-p (peek-char nil stream))
            (vector-push-extend (read-char stream) acc))
    (loop while (and (peek-char nil stream nil)
                     (digit-char-p (peek-char nil stream)))
          do (vector-push-extend (read-char stream) acc))
    (read-from-string acc)))

(defun parse-operator (stream)        
  ;; Read the operator
  (expect (find (peek-char nil stream) "+-*/")
          (read-from-string (string (read-char stream)))))

(defun parse-fraction (stream)
  (let ((acc (make-array 1 :element-type 'character :adjustable t :fill-pointer 0)))
    ;; Read a possible -
    (when (char= #\- (peek-char nil stream))
      (vector-push-extend (read-char stream) acc))
    ;; Read the number
    (expect (digit-char-p (peek-char nil stream))
            (vector-push-extend (read-char stream) acc))
    (loop while (digit-char-p (peek-char nil stream))
          do (vector-push-extend (read-char stream) acc))
    ;; read fraction
    (expect (char= #\/ (peek-char nil stream))
            (vector-push-extend (read-char stream) acc))
    ;; Read the number
    (expect (digit-char-p (peek-char nil stream))
            (vector-push-extend (read-char stream) acc))
    (loop while (and (peek-char nil stream nil)
                     (digit-char-p (peek-char nil stream)))
          do (vector-push-extend (read-char stream) acc))
    ;; CL simplifies the fraction. So don't accept it unless it's simplified.
    (let ((frac (read-from-string acc)))
      (when (string/= (prin1-to-string frac) acc)
        (error "unreduced fraction"))
      frac)))

;; (defmacro define-input-parser (name &rest fields)
;;   (with-gensyms (str parse-templ p1 stream)
;;     `(defun ,name (str)
;;        (macrolet ((,parse-templ ,(loop for f in fields collect (second f))
;;                      (let ((,stream (make-string-input-stream str)))
;;                        ;; read space
;;                        (loop while (find (peek-char nil ,stream) '(#\Space #\Tab)) do (read-char ,stream))
;;                        (let ,(loop for f in fields
;;                                    collect `(,(first f) (prog1
;;                                                             (,(second f) ,stream)
;;                                                           (loop while (find (peek-char nil ,stream) '(#\Space #\Tab)) do (read-char ,stream)))))
;;                          ;; read trailing space
;;                          (loop while (find (peek-char nil ,stream) '(#\Space #\Tab)) do (read-char ,stream))
;;                          (if (eq (peek-char nil ,stream nil ,stream) ,stream)
;;                              (list ,@(loop for f in fields collect (first f)))
;;                            (error "trailing garbage"))))))
;;          (or ,@(loop for 

(defun parse-seq (str &rest fns)
  "use each fn in fns to parse string. Eat whitespace before and
  after each fn. Return the result in a list or signal an
  error if it wasn't able to parse."
  (labels ((eat-space (stream)
             (loop while (find (peek-char nil stream) '(#\Space #\Tab)) do (read-char stream)))
           (maybe-eat-space (stream)
            (loop while (and (peek-char nil stream nil)
                             (find (peek-char nil stream) '(#\Space #\Tab))) do (read-char stream))))
    (let ((stream (make-string-input-stream str)))
      (prog1
          (loop for fn in fns
                collect (progn
                          (eat-space stream)
                          (funcall fn stream)))
        (maybe-eat-space stream)
        (unless (eq (peek-char nil stream nil stream) stream)
          (error "trailing garbage"))))))

(defun parse-arg-op-input (str)
  (or (ignore-errors (parse-seq str 'parse-number 'parse-operator))
      (ignore-errors (parse-seq str 'parse-fraction 'parse-operator))))
  
(defun parse-op-arg-input (str)
  (or (ignore-errors (parse-seq str 'parse-operator 'parse-number))
      (ignore-errors (parse-seq str 'parse-operator 'parse-fraction))))

(defun parse-box-input (str)
  (macrolet ((line (fn1 fn2)
                   `(ignore-errors (parse-seq str ,fn1 'parse-operator ,fn2))))
    (or (line 'parse-number 'parse-number)
        (line 'parse-number 'parse-fraction)
        (line 'parse-fraction 'parse-number)
        (line 'parse-fraction 'parse-fraction))))

;; (defun old-get-read-fn (box)
;;   (if *use-mouse-input*
;;       (lambda (title text fn)
;;         (let ((p (convert-coordinates (make-point 0 (view-height (box-answer-view box)))
;;                                       (box-answer-view box)
;;                                       *window*)))
;;           (read-mouse-input (point-h p) (point-v p) title text fn)))
;;       'read-template))

(defun get-read-fn (box)
  (let ((input-fn (ecase *input-method* (:mouse 'read-mouse-input) (:kbd 'read-kbd-input) (:dialog 'read-dialog-input))))
    (lambda (title text fn)
      (let* ((view (etypecase box (box (box-answer-view box)) (argument (arg-widget box))))
             (p (convert-coordinates (make-point 0 (view-height view))
                                     view
                                     *window*)))
        (funcall input-fn (point-h p) (point-v p) title text fn box)))))

(defun read-box (post-fn box)
  "Call post-fn with a list of up to 3 atoms or nil if it failed."
  (let ((fn (get-read-fn box)))
    (funcall fn "Input"
             "Fractions are written like this: 1/2, 13/8, etc."
             (lambda (input)
               (let* ((a (parse-box-input input))
                      (b (and *allow-pre-evaled-box-input*
                              (parse-number-input input)))
                      (answer (or a b)))
                 (funcall post-fn answer input))))))

(defun read-arg-op (post-fn box)
  "Call post-fn with a list of up to 3 atoms or nil if it failed."
  (let ((fn (get-read-fn box)))
    (funcall fn "Input"
             "Fractions are written like this: 1/2, 13/8, etc."
             (lambda (input)
               (let ((a (parse-arg-op-input input))
                     )
;;                  (when (and a (string/= (prin1-to-string a) input))
;;                    (setf a nil
;;                          err (format nil "unreduced fraction in: ~a" input)))
                 (funcall post-fn a input))))))

(defun read-op-arg (post-fn box)
  "Call post-fn with a list of up to 3 atoms or nil if it failed."
  (let ((fn (get-read-fn box)))
    (funcall fn "Input"
             "Fractions are written like this: 1/2, 13/8, etc."
             (lambda (input)
               (let* ((a (parse-op-arg-input input))
                      )
;;                  (when (and a (string/= (prin1-to-string a) input))
;;                    (setf a nil
;;                          err (format nil "unreduced fraction in: ~a" input)))
                 (funcall post-fn a input))))))

;; (let* ((stream (make-string-input-stream input))
;;                                            (tokens (list (ignore-errors (read stream))
;;                                                          (ignore-errors (read stream))
;;                                                          (ignore-errors (read stream)))))
;;                                       (when (and (numberp (first tokens))
;;                                                  (find (second tokens) '(+ - * /))
;;                                                  (numberp (third tokens)))
;;                                         tokens))

(defun parse-number-input (input)
  (let ((n (ignore-errors 
             (multiple-value-bind (n p) (read-from-string input)
               (assert (= p (length input)))
               n))))
    ;;(format t "~a~%" n)
    ;; CL reduces fractions for us. don't let the user enter a unreduced one.
    (when (and (numberp n)
               (string= (prin1-to-string n) input))
      n)))

(defun parse-operator-input (input)
  (let ((op (ignore-errors 
             (multiple-value-bind (op p) (read-from-string input)
               (assert (= p (length input)))
               op))))
    ;;(format t "~a~%" n)
    (when (find op '(+ - * /))
      op)))

(defun read-number (post-fn arg)
  "call POST-FN with a symbol or nil if it failed."
  (let ((fn (get-read-fn arg)))
    (funcall fn "Input"
             "Fractions are written like this: 1/2, 13/8, etc."
             (lambda (input)
               (funcall post-fn (parse-number-input input) input)))))

(defun read-operator (post-fn arg)
  "call post-fn with a symbol or nil if it failed."
  (let ((fn (get-read-fn arg)))
    (funcall fn "Input"
             "Fractions are written like this: 1/2, 13/8, etc."
             (lambda (input)
                 (funcall post-fn (parse-operator-input input) input)))))

(defmethod window-close-event-handler :after ((window dialog))
  (handle-window-close window))

(defun handle-window-close (window)
  (history-add-close-dialog)
  (when (eq window *input-window*)
    (setf *input-window* nil
          *fill-in-object* nil
          *input-text* nil))
  ;; for finegrained hints
  (update-hints))

(defun type-in-dialog (dialog char)
  (let ((view (find-if (lambda (d)
                         (typep d 'editable-text-dialog-item))
                       (subviews dialog))))
    (cond ((graphic-char-p char) (ed-insert-char view char))
          ((char= char #\backspace) (ed-rubout-char view)))
    (view-draw-contents view)))

(defun submit-dialog (dialog)
  (if (typep dialog 'ccl::string-dialog)
      (let ((view (find-if (lambda (d)
                             (typep d 'default-button-dialog-item))
                           (subviews dialog))))
        (press-button view))
      (input-submit-it nil)))


;;;;;; input in the MRI 

(defclass input-window (view)
  ((parent :initarg :parent :accessor input-window-parent)))

(defparameter +input-cursor+ "_")
(defvar *input-post-fn* nil)

(defparameter +input-window-width+ 175)
(defparameter +input-window-height+ 147)

(defun input-submit-it (click-p)
  (let ((txt (string-butlast (dialog-item-text *input-text*))))
    (if (must-supply-user-hints-p)
        (set-message-text +fill-in-user-hint-text+)
        (progn
          ;;(window-close *input-window*)
          (remove-subviews (view-container *input-window*) *input-window*)
          ;; restore the operation buttons
          ;; (unhide-view (tutor-buttons *tutor*))
          (setf *input-window* nil
                *input-text* nil
                *fill-in-object* nil)
          (if click-p
              (history-input-click-button 'submit)
              (history-add 'input-type 'submit))
          (funcall *input-post-fn* txt)))))

(defun input-add-input (str)
  (set-dialog-item-text *input-text* (concatenate 'string (string-butlast (dialog-item-text *input-text*)) str +input-cursor+)))

;; (defun input-clear ()
;;   (set-dialog-item-text *input-text* +input-cursor+)
;;   (history-input-click-button 'clear))

(defun input-backspace ()
  (labels ((del-spaces (s)
             (loop for i from (length s) downto 1
                while (char= (char s (1- i)) #\Space)
                finally return (subseq s 0 i)))
           (del-not-space (s)
             (if (and (> (length s) 0)
                      (char/= (char s (1- (length s))) #\Space))
                 (subseq s 0 (1- (length s)))
                 s)))
    ;; delete whitespace leading up to the char then the char then whitespace
    (let ((s (string-butlast (dialog-item-text *input-text*))))
      (when (> (length s) 0)
        (set-dialog-item-text *input-text* (concatenate 'string
                                                        (if (char= (char s (1- (length s))) #\Space)
                                                            (del-spaces (del-not-space (del-spaces s)))
                                                            (del-not-space s)) +input-cursor+)))
      (history-add 'click-button 'backspace))))

(defun handle-kbd-input (ch)
  (case ch
    (#\Backspace
     (let ((s (string-butlast (dialog-item-text *input-text*))))
       (when (> (length s) 0)
         (set-dialog-item-text *input-text* (concatenate 'string
                                                         (string-butlast s)
                                                         +input-cursor+)))
       (history-log-input 'backspace)))
    ((#\Return #\Enter)
     (input-submit-it nil))
    (t (if (find ch "1234567890/+-* ")
           (progn
             (input-add-input (string ch))
             (history-log-input ch))
           (history-add-keypress ch)))))

(defun read-kbd-input (x y title text post-fn thing &optional (window *window*))
  (declare (ignore text title))
  (when *input-window*
    (remove-subviews (view-container *input-window*) *input-window*)
    (history-input-close))
  (history-input-open)
  (setf *input-window* (make-instance 'input-window
                                      :parent thing
                                      :view-font '("Monaco" 14)
                                      :view-position (make-point x y)
                                      :view-container window
                                      :view-size (make-point +input-window-width+ 30)))
  (let ((views (create-layout `(:vbox :border t
                                      :contents ((:hbox :contents ((:button :width 37 :text "^" :action ,(button-fn (input-submit-it t)))
                                                                   (:text :text "" :width 4 :font ,*default-font*)
                                                                   (:text :text ,+input-cursor+ :name text :font ,*default-font*)))))
                              *input-window*)))
    (invalidate-view *input-window*)
    (setf *input-text* (cdr (assoc 'text views))
          *input-post-fn* post-fn)
    ;; for finegrained hints
    (update-hints)))

(defun read-mouse-input (x y title text post-fn thing &optional (window *window*))
  (declare (ignore text title))
  (macrolet ((input (s)
               (with-gensyms (foo)
                 `(lambda (,foo) 
                    (declare (ignore ,foo))
                    (input-add-input ,s)
                    (history-input-click-button ,s)))))
    ;;     (hide-view (tutor-buttons *tutor*))
    (when *input-window*
      (remove-subviews (view-container *input-window*) *input-window*)
      (history-input-close))
    (history-input-open)
    (setf *input-window* (make-instance 'input-window
                                        :parent thing
                                        :view-font '("Monaco" 14)
                                        :view-position (make-point x y)
                                        :view-container window
                                        :view-size (make-point +input-window-width+ +input-window-height+)))
    (let ((views (create-layout `(:vbox :border t
                                        :contents ((:hbox :contents ((:button :width 37 :text "^" :action ,(button-fn (input-submit-it t)))
                                                                     (:text :text "" :width 4 :font ,*default-font*)
                                                                     (:text :text ,+input-cursor+ :name text :font ,*default-font*)))
                                                   (:hbox :contents ((:button :text "7" :action ,(input "7"))
                                                                     (:button :text "8" :action ,(input "8"))
                                                                     (:button :text "9" :action ,(input "9"))
                                                                     (:button :text "+" :action ,(input "+"))))
                                                   (:hbox :contents ((:button :text "4" :action ,(input "4"))
                                                                     (:button :text "5" :action ,(input "5"))
                                                                     (:button :text "6" :action ,(input "6"))
                                                                     (:button :text "-" :action ,(input "-"))))
                                                   (:hbox :contents ((:button :text "1" :action ,(input "1"))
                                                                     (:button :text "2" :action ,(input "2"))
                                                                     (:button :text "3" :action ,(input "3"))
                                                                     (:button :text "*" :action ,(input "*"))))
                                                   (:hbox :contents ((:button :text "0" :action ,(input "0"))
                                                                     (:button :text "<-" :action ,(button-fn (input-backspace)))
                                                                     (:button :text "" :action ,(input "/"))
                                                                     (:button :text "/" :action ,(input " / "))))
                                                   ))
                                *input-window*)))
      (invalidate-view *input-window*)
      (setf *input-text* (cdr (assoc 'text views))
            *input-post-fn* post-fn)
      ;; for finegrained hints
      (update-hints))))


;;; paragraph modal dialog

(defclass paragraph-window (window)
  ())

(defmethod view-key-event-handler :around ((view paragraph-window) ch)
  (if (find ch '(#\Newline #\Enter #\Return) :test 'char=)
      (return-from-modal-dialog (dialog-item-text (view-named 'edit view)))
      (call-next-method)))

(defun get-paragraph-from-user (width height &key (x 50) (y 50) (title "Input Paragraph") (pad 5) (okay t) (text ""))
  (let ((w (make-instance 'paragraph-window 
                          :window-title title
                          :window-type :document
                          :view-size (make-point width height)
                          :view-position (make-point x y)
                          :view-subviews
                          (list*
                           (make-instance 'editable-text-dialog-item
                                          :view-position (make-point pad pad)
                                          :wrap-p t
                                          :view-font *default-font*
                                          :view-nick-name 'edit
                                          :view-size (make-point (- width pad pad) (- height (if okay 40 pad) pad))
                                          :dialog-item-text text
                                          :allow-returns nil)
                           (when okay
                             (make-instance 'button-dialog-item
                                            :view-position (if okay 
                                                               (make-point pad (- height 30))
                                                               (make-point pad pad))
                                            :default-button t
                                            :dialog-item-text "Okay"
                                            :dialog-item-action (lambda (b) (return-from-modal-dialog (dialog-item-text (view-named 'edit (view-container b)))))))))))
    (modal-dialog w)))
