(in-package :tutor)

(defun make-problem-template (instructions template answer nright &rest constraints)
  (list 'learn
        (make-instance 'problem-type
                       :instructions instructions
                       :template template
                       :constraints constraints
                       :nright nright
                       :template-answer answer)))

(defparameter *negative-chance* 5)
(defun fill-problem-template (template answer hint)
  "Create a problem from a template"
  (let (bindings)
    (labels ((s= (s1 s2)
               (let ((a (symbol-name s1))
                     (b (symbol-name s2)))
                 (string= (subseq a 0 (min (length a) (length b))) (subseq b 0 (min (length a) (length b))))))
             (doit (templ)
               (cond
                 ((listp templ)
                  (mapcar #'doit templ))
                 ((s= templ :n)
                  (let ((r (1+ (random 9))))
                    (when (zerop (random *negative-chance*)) (setf r (* -1 r)))
                    (push (list templ r) bindings)
                    r))
                 ((s= templ :nn)
                  (let ((r (1+ (random 99))))
                    (when (zerop (random *negative-chance*)) (setf r (* -1 r)))
                    (push (list templ r) bindings)
                    r))
                 ((s= templ :a)
                  (let ((r (if (zerop (random 2)) '- '+)))
                    (push (list templ r) bindings)
                    r))
                 ((s= templ :m)
                  (let ((r (if (zerop (random 2)) '* '/)))
                    (push (list templ r) bindings)
                    r))
                 ((s= templ :o)
                  (let ((r (nth (random 4) '(* - / +))))
                    (push (list templ r) bindings)
                    r))
                 (t templ))))
      (let ((a (doit template))
            (b (compile-dest-pattern bindings answer)))
        (make-problem a (perms b) 1 hint)))))

(defun make-problems (instructions hint-modifiers problems)
  (list 'problem
        (make-instance 'problem-set
                       :instructions instructions
                       :hint-modifiers hint-modifiers
                       :problems problems)))

(defun make-problem (p a points &optional hint hint-modifiers)
  (make-instance 'problem
		 :eqn p
		 :answer a
                 :points points
		 :auto-hint hint
                 :hint-modifiers (if (listp hint-modifiers) 
                                     hint-modifiers
                                     (list hint-modifiers))))

(defun make-text (&rest txt)
  (cons 'text txt))

(defun glob-perms (op a b)
  "A and B are lists and OP is an operator. Return a list of all
combinations of OP the elements of A and B."
  (apply #'nconc (mapcar (lambda (d)
                           (mapcar (lambda (e)
                                     (list op e d))
                                   a))
                         b)))
  
(defun perms (base)
  "Return all equivalent permutations of a problem."
  (cond ((or (typep base 'number)
	     (typep base 'symbol))
	 (list base))
	((= (length base) 1)
	 (mapcar #'list (perms (first base))))
	((= (length base) 3)
	 (let ((a (apply #'op-answers base)))
	   (apply #'nconc (mapcar (lambda (n)
				    (let ((other (when (and (find (first n) '(- +))
                                                            (numberp (second n))
                                                            (listp (third n))
                                                            (let ((m (third n)))
                                                              (and (find (first m) '(* /))
                                                                   (symbolp (third m))
                                                                   (numberp (second m)))))
                                                   ;; Handle (- N (* M X))
                                                   (let ((m (third n)))
                                                     (glob-perms (invert-op (first n)) (list (second n)) (perms `(,(first m) ,(- (second m)) ,(third m)))))))
                                          (other2 (when (and (find (first n) '(- +))
                                                             (numberp (second n))
                                                             (symbolp (third n)))
                                                    (if (eq (first n) '-)
                                                        (glob-perms '+ (list (second n)) (perms `(* -1 ,(third n))))
                                                      (glob-perms '- (list (second n)) (perms `(* -1 ,(third n)))))))
                                           (b (perms (second n)))
                                           (c (perms (third n))))
                                      (nconc other other2 (glob-perms (first n) b c))))
				  a))))
	(t (error "unknown type ~a" (type-of base)))))

(defun algebraically-equivalent-p (e1 e2)
  "Determine if e1 and e2 are algebraically equivalent. This uses
a couple rules to cover most of the cases found in the tutor."
  (macrolet ((special-case (val match recode)
               (with-gensyms (i)
                 `(or
                   (let ((,val e1))
                     (and (simple-match-2 ,val ,match)
                          (loop for ,i in (apply 'op-answers ,recode)
                             thereis (trivially-equivalent ,i e2))))
                   (let ((,val e2))
                     (and (simple-match-2 ,val ,match)
                          (loop for ,i in (apply 'op-answers ,recode)
                             thereis (trivially-equivalent e1 ,i))))))))
    (labels ((trivially-equivalent (a b)
               (if (and (listp a) (listp b))
                   (loop for i in a for j in b
                      always (algebraically-equivalent-p i j))
                   (eql e1 e2)))
             (pre-evalled-equivalent (a b)
               (and *allow-pre-evaled-box-input*
                    (numberp a) 
                    (listp b)
                    (= (length b) 3)
                    (simple-match-2 b '(:o :n :n))
                    (= a (apply (first b) (cdr b))))))
      (cond 
        ((pre-evalled-equivalent e1 e2) t)
        ((pre-evalled-equivalent e2 e1) t)
        ((or (numberp e1) (symbolp e1)
             (numberp e2) (symbolp e2))
         (eql e1 e2))
        ((and (listp e1) (listp e2))
         (if (= (length e1) 1)
             ;; answer/number box
             (and (= (length e2) 1)
                  (algebraically-equivalent-p (car e1) (car e2)))
             ;; arg box
             (or 
              (trivially-equivalent e1 e2)
              ;; less trivial cases
              (loop for i in (apply 'op-answers e2) 
                 thereis (trivially-equivalent e1 i))
              ;; special cases
              (special-case a '(- :s (:m :n :s))
                            `(+ ,(second a) (,(first (third a)) ,(- (second (third a))) ,(third (third a)))))
              (special-case a '(:a :n :x)
                            `(,(invert-op (first a)) ,(second a) (* -1 ,(third a))))
              (special-case a '(:a :s (:a :n :n))
                            `(,(invert-op (first a)) ,(second a) (,(first (third a)) ,(* -1 (second (third a))) ,(* -1 (third (third a))))))
              (special-case a '(:a :s (:a :n :s))
                            `(,(invert-op (first a)) ,(second a) (,(invert-op (first (third a))) ,(* -1 (second (third a))) ,(third (third a))))))))
        (t (error "badly formatted equation ~a ~a" e1 e2))))))
         
(defun problem (prob answer points &optional hint hint-modifiers)
  (make-problem prob (perms answer) points hint hint-modifiers))

(defun make-docs (title &rest paragraphs)
  "Produce a list markup representation of the title and paragraphs."
  (setf paragraphs (flatten paragraphs))
  (cons (list :title title)
        (mapcar (lambda (p)
                  (list :text p))
                paragraphs)))

(defun pick-prob (p1 p2)
  (if *simplified-ui*
      p2
    p1))

(defun even-or-odd-problems (exclusions valids a1 a2 b1 b2)
  (if *odd-problems*
      (values exclusions valids a1 a2)
    (values exclusions valids b1 b2)))

(defun even-odd-or-alternative-problems (exclusions valids e1 e2 o1 o2 a1 a2)
  (cond ((eq *odd-problems* nil)
         (values exclusions valids o1 o2))
        ((eq *odd-problems* :alternative)
         (values exclusions valids a1 a2))
        (t (values exclusions valids e1 e2))))

(defun gen-problems (section)
  "Return a list of problems for the given section. The first ret
val is a list of practice problems. the second is the list of
problems."
  (ecase section
    (ch1-s1
     (even-odd-or-alternative-problems
      '(:no-canonicalize)
      '(evaluate)
      (list (problem '(* (+ 5 7) 8) '(96) 2 t))
      (list
       (problem '(- 9 (* 4 2)) '(1) 2)
       (problem '(* (- 9 4) 2) '(10) 2)
       (problem '(- (* 8 7) 13) '(43) 2)
       (problem '(* 9 (+ 4 2)) '(54) 2)
       (pick-prob
        (problem '(* (/ 36 2) 6) '(108) 2)
        (problem '(* (/ 18 2) 6) '(54) 2))
       (problem '(- (- 16 (* 3 2)) 4) '(6) 3)
       (problem '(- 16 (- (* 3 2) 4)) '(14) 3)
       (pick-prob
        (problem '(* (+ 10 6) (- 11 3)) '(128) 3)
        (problem '(* (+ 10 6) (- 11 7)) '(64) 3))
       (problem '(+ (* 6 (- 11 3)) 10) '(58) 3)
       (problem '(/ 100 (- (/ 20 5) 3)) '(100) 3)
       (problem '(/ 100 (/ 20 (- 5 3))) '(10) 3)
       (problem '(- (/ (- 63 8) (+ 3 8)) 2) '(3) 4)
       (problem '(* 5 (/ (- 19 7) (+ 5 1))) '(10) 4))
      ;; odd problems
      (list (problem '(+ 5 (* 7 8)) '(61) 2 t))
      (list
       (problem '(* (+ 4 9) 5) '(65) 2)
       (problem '(+ (* 9 5) 4) '(49) 2)
       (problem '(- (* 5 2) 11) '(-1) 2)
       (problem '(* (- 11 5) 2) '(12) 2)
       (problem '(- (* 7 6) 19) '(23) 2)
       (problem '(* (+ 2 5) 8) '(56) 2)
       (problem '(* (/ 54 3) 2) '(36) 2)
       (problem '(/ 54 (* 3 2)) '(9) 2)
       (problem '(- (- 17 (* 4 3)) 2) '(3) 3)
       (problem '(* (+ 14 6) (- 9 2)) '(140) 3)
       (problem '(+ 14 (* 6 (- 9 2))) '(56) 3)
       (pick-prob
        (problem '(- (/ 800 (/ 50 2)) 7) '(25) 3)
        (problem '(- (/ 200 (/ 50 2)) 7) '(1) 3))
       (pick-prob
        (problem '(- (/ (/ 800 50) 2) 7) '(1) 3)
        (problem '(- (/ (/ 200 50) 2) 7) '(-5) 3))
       (problem '(- 19 (/ (- 25 7) (+ 5 4))) '(17) 4)
       (problem '(/ 28 (/ (* 8 6) (+ 5 7))) '(7) 4))
      ;; 3rd set 'alternative' problems
      (list (problem '(* (+ 4 7) 8) '(88) 2 t))
      (list
       (problem '(- 11 (* 5 2)) '(1) 2)
       (problem '(* (- 9 3) 3) '(18) 2)
       (problem '(- (* 8 6) 11) '(37) 2)
       (problem '(* 7 (+ 3 4)) '(49) 2)
       (pick-prob
        (problem '(* (/ 36 2) 6) '(108) 2)
        (problem '(* (/ 16 2) 7) '(56) 2))
       (problem '(- (- 16 (* 4 2)) 4) '(4) 3)
       (problem '(- 16 (- (* 4 2) 4)) '(12) 3)
       (pick-prob
        (problem '(* (+ 10 6) (- 11 3)) '(128) 3)
        (problem '(* (+ 12 4) (- 9 5)) '(64) 3))
       (problem '(+ (* 6 (- 10 3)) 8) '(50) 3)
       (problem '(/ 80 (- (/ 30 6) 4)) '(80) 3)
       (problem '(/ 50 (/ 40 (- 7 3))) '(5) 3)
       (problem '(- (/ (- 44 8) (+ 3 3)) 2) '(4) 4)
       (problem '(* 8 (/ (- 22 7) (+ 2 3))) '(24) 4))))
  (Ch1-s2
      (even-or-odd-problems
       '(:no-canonicalize)
       '(evaluate)
       (list 
        (problem '(+ 5 (4)) '(9) 2 t))
       (list
        (problem '(+ (93) 5) '(98) 2)
        (problem '(- (/ 24 (12)) 1) '(1) 3)
        (problem '(* (- (37) 19) 2) '(36) 3)
        (problem '(* (- (31) 31) 4) '(0) 3)
        (problem '(* 7 (11)) '(77) 2)
        (problem '(/ 40 (5)) '(8) 2)
        (problem '(/ (+ 5 (95)) 2) '(50) 3)
        (problem '(/ (+ (0) 7) 3) '(7/3) 3))
       ;; even problems
       (list 
        (problem '(+ 5 (6)) '(11) 2 t))
       (list
        (problem '(+ (8) 5) '(13) 2)
        (problem '(+ (/ 72 (9)) 5) '(13) 3)
        (pick-prob
         (problem '(* (- (80) 31) 4) '(196) 3)
         (problem '(* (- (80) 31) 2) '(98) 3))
        (problem '(* 9 (13)) '(117) 2)
        (problem '(/ 30 (90)) '(1/3) 2)
        (problem '(/ 40 (60)) '(2/3) 2)
        (problem '(/ (+ (5) 7) 3) '(4) 3))))
  (ch1-s7
   (even-or-odd-problems
    '(:no-canonicalize)
    '(evaluate invert unwind)
    (list
     (problem '(= (+ x 3) 8) '(= x 5) 2 t))
    (list
     (problem '(= (- x 10) 17) '(= x 27) 2)
     (problem '(= (+ x 4) 13) '(= x 9) 2)
     (problem '(= (* 5 x) 90) '(= x 18) 2)
     (pick-prob
      (problem '(= (/ x 3) 29) '(= x 87) 2)
      (problem '(= (/ x 3) 21) '(= x 63) 2))
     (problem '(= (- x 16) 7) '(= x 23) 2)
     (problem '(= (- x 32) 95) '(= x 127) 2)
     (pick-prob
      (problem '(= (+ x 58) 74) '(= x 16) 2)
      (problem '(= (+ x 54) 74) '(= x 20) 2))
     (pick-prob
      (problem '(= (+ x 91) 27) '(= x -64) 2)
      (problem '(= (+ x 91) 87) '(= x -4) 2))
     (pick-prob
      (problem '(= (* 12 x) 132) '(= x 11) 2)
      (problem '(= (* 12 x) 48) '(= x 4) 2))
     (pick-prob
      (problem '(= (/ x 7) 68) '(= x 476) 2)
      (problem '(= (/ x 7) 5) '(= x 35) 2))
     (problem '(= (/ x 10) 100) '(= x 1000) 2)
     (problem '(= (+ x 5) 4) '(= x -1) 2)
     (problem '(= (- x 94) 362) '(= x 456) 2)
     (problem '(= (+ x 1984) 2001) '(= x 17) 2)
     (problem '(= (/ x 3) 51) '(= x 153) 2)
     (problem '(= (* 3 x) 51) '(= x 17) 2)
     (problem '(= (* 17 x) 51) '(= x 3) 2)
     (pick-prob
      (problem '(= (/ x 17) 51) '(= x 867) 2)
      (problem '(= (/ x 5) 50) '(= x 250) 2))
     (problem '(= (* 6 x) 2) '(= x 1/3) 2)
     (problem '(= (* 2/3 x) 6) '(= x 9) 2)
     (problem '(= (* 3/5 x) 30) '(= x 50) 2)
     (pick-prob
      (problem '(= (* 7/2 x) 28) '(= x 8) 2)
      (problem '(= (* 7 x) 28) '(= x 4) 2))
     (problem '(= (+ x 2/3) 5/3) '(= x 1) 2)
     (problem '(= (- x 3/4) 7/4) '(= x 5/2) 2))
    ;; even problems
    (list (problem '(= (+ x 5) 17) '(= x 12) 2 t))
    (list
     (problem '(= (- x 8) 13) '(= x 21) 2)
     (problem '(= (+ x 9) 21) '(= x 12) 2)
     (pick-prob
      (problem '(= (* 6 x) 72) '(= x 12) 2)
      (problem '(= (* 6 x) 66) '(= x 11) 2))
     (problem '(= (* 1/5 x) 23) '(= x 115) 2)
     (problem '(= (- x 17) 8) '(= x 25) 2)
     (problem '(= (- x 45) 89) '(= x 134) 2)
     (problem '(= (+ x 43) 91) '(= x 48) 2)
     (pick-prob
      (problem '(= (+ x 79) 422) '(= x 343) 2)
      (problem '(= (+ x 80) 422) '(= x 342) 2))
     (pick-prob
      (problem '(= (* 15 x) 210) '(= x 14) 2)
      (problem '(= (* 15 x) 300) '(= x 20) 2))
     (problem '(= (* 4 x) 100) '(= x 25) 2)
     (pick-prob
      (problem '(= (* 1/9 x) 58) '(= x 522) 2)
      (problem '(= (* 1/9 x) 50) '(= x 450) 2))
     (problem '(= (* 1/10 x) 1000) '(= x 10000) 2)
     (problem '(= (+ x 7) 5) '(= x -2) 2)
     (problem '(= (- x 61) 365) '(= x 426) 2)
     (problem '(= (+ x 1066) 1776) '(= x 710) 2)
     (pick-prob
      (problem '(= (* 1/3 x) 57) '(= x 171) 2)
      (problem '(= (* 1/3 x) 55) '(= x 165) 2))
     (pick-prob
      (problem '(= (* 3 x) 57) '(= x 19) 2)
      (problem '(= (* 3 x) 54) '(= x 18) 2))
     (pick-prob
      (problem '(= (* 19 x) 57) '(= x 3) 2)
      (problem '(= (* 18 x) 54) '(= x 3) 2))
     (pick-prob
      (problem '(= (* 1/19 x) 57) '(= x 1083) 2)
      (problem '(= (* 1/19 x) 50) '(= x 950) 2))
     (problem '(= (* 8 x) 4) '(= x 1/2) 2)
     (problem '(= (* 3/2 x) 6) '(= x 4) 2)
     (problem '(= (+ x 3/5) 13/5) '(= x 2) 2)
     (problem '(= (- x 2/3) 14/3) '(= x 16/3) 2)
     )))
  (ch2-s1
     (values
      '(:no-canonicalize :verbose-minus)
      '(evaluate)
      (list (list
       (problem '(+ 4 -3) '(1) 1 t)
       (problem '(+ -3 -4) '(-7) 1 t)))
      (list
       (problem '(+ 5 -2) '(3) 1)
       (problem '(+ 6 -4) '(2) 1)
       (problem '(+ 2 -5) '(-3) 1)
       (problem '(+ 13 -8) '(5) 1)
       (problem '(+ 3 -7) '(-4) 1)
       (problem '(+ -3 -2) '(-5) 1)
       (problem '(+ -13 -4) '(-17) 1)
       (problem '(+ -4 -5) '(-9) 1)
       (problem '(+ -8 -9) '(-17) 1)
       (problem '(+ 0 -3) '(-3) 1)
       (problem '(+ -3 5) '(2) 1)
       (problem '(+ -4 1) '(-3) 1)
       (problem '(+ -13 15) '(2) 1)
       (problem '(+ -27 1) '(-26) 1)
       (problem '(+ -11 35) '(24) 1))))
    (ch2-s3
     (values
      '(:no-canonicalize :verbose-minus)
      '(evaluate)
      (list
       (problem '(- -4 3) '(-7) 1 t))
      (list
       (problem '(- 35 200) '(-165) 1)
       (problem '(- 14 -17) '(31) 1)
       (problem '(- -11 -2) '(-9) 1)
       (problem '(- -99 59) '(-158) 1)
       (problem '(- -241 -42) '(-199) 1)
       (problem '(- 158 -115) '(273) 1)
       (problem '(- 5/7 -2/7) '(1) 1)
       (problem '(- 0 -15) '(15) 1)
       (problem '(- 154 (- 43 28)) '(139) 2)
       (problem '(- 314 (- 58 70)) '(326) 2)
       (problem '(- (- 33 45) (- 64 78)) '(2) 3)
       (problem '(- (- 13 5) (- -2 6)) '(16) 3)
       (problem '(+ (- 1066 1492) 1984) '(1558) 2)
       (problem '(- (- 8 -3) (- 4 -6)) '(1) 3)
       (problem '(+ (+ (- 1 2) (- 4 8)) 16) '(11) 4)
       (problem '(- (+ (- (- 45 7) 58) 11) 21) '(-30) 4)
       (problem '(- (+ (- (+ (- 19 2) 23) 54) 34) 29) '(-9) 5)
       (problem '(- (+ (- (+ (- 200 195) 71) 66) 18) 29) '(-1) 5))))
    (ch2-s4
     (values
      '(:no-canonicalize :verbose-minus)
      '(evaluate)
      (list (list
       (problem '(* -3 -7) '(21) 1 t)
       (problem '(* -5 8) '(-40) 1 t)))
      (list
       (problem '(* 10 -7) '(-70) 1)
       (problem '(* -5 -9) '(45) 1)
       (problem '(* -6 8) '(-48) 1)
       (problem '(* -2/3 6) '(-4) 1)
       (problem '(* -72 1) '(-72) 1)
       (problem '(* -1 -200) '(200) 1)
       (problem '(* -53 0) '(0) 1)
       (problem '(* (* -3 -5) -7) '(-105) 2)
       (problem '(* (* -2 5) -8) '(80) 2)
       (problem '(* (* 4 -7) 2) '(-56) 2)
       (problem '(* (* (* -1 5) -3) -2) '(-30) 3))))
    (ch2-s5
     (values
      '(:no-canonicalize :verbose-minus)
      '(evaluate)
      (list (list 
       (problem '(/ -54 6) '(-9) 1 t)
       (problem '(/ -54 -3) '(18) 1 t)))
      (list
       (problem '(/ -21 7) '(-3) 1)
       (problem '(/ 21 -7) '(-3) 1)
       (problem '(/ -21 -7) '(3) 1)
       ;; 7 uses notation we don't have
       ;; 9 uses notation we don't have
       ;; 11 uses notation we don't have
       (problem '(/ -26 2) '(-13) 1)
       (problem '(/ 51 -3) '(-17) 1)
       (problem '(/ -55 -11) '(5) 1)
       (problem '(/ 1000 -10) '(-100) 1)
       (problem '(/ (- 7 19) -3) '(4) 2)
       (problem '(/ (- 13 37) -2) '(12) 2)
       (problem '(/ 45 (- 5 14)) '(-5) 2)
       (problem '(/ (- 32 4) (- 8 15)) '(-4) 3)
       (problem '(/ (* 8 -3) -6) '(4) 2)
       (problem '(/ (* -3 10) 5) '(-6) 2)
       (problem '(/ 10 -1) '(-10) 1)
       (problem '(/ -10 -1) '(10) 1)
       (problem '(/ 0 10) '(0) 1)
       (problem '(/ 0 -10) '(0) 1)
       ;; 39 is division by zero (DbZ)
       (problem '(/ 15 (5)) '(3) 2)
       (problem '(/ 15 (-3)) '(-5) 2)
       ;; 41.c is DbZ
       (problem '(/ (5) -30) '(-1/6) 2)
       (problem '(/ (-3) -30) '(1/10) 2)
       ;; 43.c is DbZ
       (problem '(/ (+ (-2) 5) (+ (-2) 3)) '(3) 3)
       (problem '(/ (+ (5) 5) (+ (5) 3)) '(5/4) 3)
       ;; 47.c is DbZ
       (problem '(/ (- 5 (2)) (- 3 (2))) '(3) 3)
       (problem '(/ (- 5 (5)) (- 3 (5))) '(0) 3)
       (problem '(/ (- 5 (-3)) (- 3 (-3))) '(4/3) 3)
       )))
    (ch2-s6
     (even-or-odd-problems
      '(:no-canonicalize)
      '(evaluate combine collect)
      (list
       (problem '(+ 3 (+ (* 2 x) 7)) '((+ (* 2 x) 10)) 2 t))
      (list
       (problem '(+ (+ 4 x) 3) '((+ x 7)) 2)
       (problem '(+ (+ 6 x) 7) '((+ x 13)) 2)
       (problem '(- (+ 5 x) 3) '((+ x 2)) 2)
       (problem '(+ (+ -8 x) 4) '((+ x -4)) 2)
       (problem '(+ (- 15 x) 9) '((- 24 x)) 2)
       (problem '(- (- 6 x) 37) '((- -31 x)) 2)
       ;; not sure what to change or if i should change this one
       (pick-prob
        (problem '(- (- -19 x) 22) '((- -41 x)) 2)
        (problem '(- (- -19 x) 22) '((- -41 x)) 2))
       (problem '(+ (+ 9 (* 7 x)) 3) '((+ 12 (* 7 x))) 2)
       (problem '(+ (- -18 (* 3 x)) 7) '((- -11 (* 3 x) )) 2)
       (problem '(* (* 5 x) 7) '((* 35 x)) 2)
       (problem '(* (* 3 x) -8) '((* -24 x)) 2)
       (problem '(* (* 20 x) 1/5) '((* 4 x)) 2)
       (problem '(* -1/8 (* 56 x)) '((* -7 x)) 2)
       (problem '(* (* 9 x) 1/9) '((* 1 x)) 2)
       (problem '(/ (* 54 x) 9) '((* 6 x)) 2)
       (problem '(/ (* -88 x) 11) '((* -8 x)) 2)
       (problem '(/ (* 1000 x) -10) '((* -100 x)) 2)
       (problem '(/ (* -37 x) -1) '((* 37 x)) 2)
       (problem '(/ (* -19 x) -19) '((* 1 x)) 2))
      (list (problem '(* (* 12 x) 1/4) '((* x 3)) 2 t))
      (list
       (problem '(+ (+ 5 x) 4) '((+ x 9)) 2)
       (problem '(+ (+ 8 x) 13) '((+ 21 x)) 2)
       (problem '(- (+ 7 x) 9) '((- x 2)) 2) 
       (problem '(+ (+ -3 x) 11) '((+ x 8)) 2)
       (problem '(+ (- 16 x) 5) '((- 21 x)) 2)
       (problem '(- (- 12 x) 15) '((- -3 x)) 2)
       (problem '(- (- -2 x) 71) '((- -73 x)) 2)
       (problem '(+ (- -13 x) 13) '((- 0 x)) 2)
       (problem '(+ (+ 14 (* 3 x)) 21) '((+ 35 (* 3 x))) 2)
       (problem '(- (- 8 (* 5 x)) 13) '((- -5 (* 5 x))) 2)
       (problem '(* (* 7 x) 9) '((* x 63)) 2)
       (problem '(* (* 5 x) -13) '((* x -65)) 2)
       (problem '(* (* 30 x) 1/6) '((* x 5)) 2)
       (problem '(* -1/3 (* 21 x)) '((* x -7)) 2)
       (problem '(* (* 1/3 x) 3) '((* x 1)) 2)
       (problem '(/ (* 86 x) 2) '((* x 43)) 2)
       (problem '(/ (* -18 x) 9) '((* x -2)) 2)
       (problem '(/ (* 100 x) -5) '((* x -20)) 2)
       (problem '(/ (* -1 x) -1) '((* x 1)) 2)
;;        (problem '(/ (* -25 x) -25) '((* x 1)) 2)
       )))
    (ch2-s7
     (even-or-odd-problems
      '(:no-canonicalize)
      '(evaluate invert unwind)
      (list
       (problem '(= (+ (* 2 x) 3) 19) '(= x 8) 4 t)
       (problem '(= (- 10 x) 2) '(= x 8) 2 t))
      (list
       (problem '(= (+ (* 1/3 x) 5) 7) '(= x 6) 4)
       (problem '(= (- (* 1/2 x) 8) 3) '(= x 22) 4)
       (problem '(= (+ (* 5 x) 4) 39) '(= x 7) 4)
       (problem '(= (- (* 3 x) 7) 26) '(= x 11) 4)
       (problem '(= (+ (* -1/6 x) 5) 8) '(= x -18) 4)
       (pick-prob
        (problem '(= (- (* -1/4 x) 9) 17) '(= x -104) 4)
        (problem '(= (- (* -1/4 x) 9) 16) '(= x -100) 4))
       (problem '(= (- 13 (* 4 x)) 25) '(= x -3) 4)
       (pick-prob
        (problem '(= (- 8 (* 1/7 x)) -19) '(= x 189) 4)
        (problem '(= (- 8 (* 1/7 x)) -9) '(= x 119) 4))
       (problem '(= 53 (+ (* 5 x) 11)) '(= x 42/5) 4)
       (problem '(= (- 17 x) 25) '(= x -8) 2)
       (problem '(= 71 (- 4 x)) '(= x -67) 2)
       (problem '(= (+ (* 2/3 x) 6) 18) '(= x 18) 4)
       (problem '(= (+ -21 (* 4 x)) 10) '(= x 31/4) 4)
       (problem '(= (- -4/3 x) -1/3) '(= x -1) 2))
      (list (problem '(= (+ (* 3/2 x) 4) -9) '(= x -26/3) 4 t)
            (problem '(= (- 5 x) -7) '(= x 12) 2 t))
      (list
       (problem '(= (+ (* 1/5 x) 3) 9) '(= x 30) 4)
       (problem '(= (- (* 1/4 x) 7) 2) '(= x 36) 4)
       (problem '(= (+ (* 6 x) 5) 53) '(= x 8) 4)
       (problem '(= (- (* 2 x) 13) 95) '(= x 54) 4)
       (problem '(= (+ (* -1/3 x) 6) 10) '(= x -12) 4)
       (problem '(= (- (* -1/2 x) 7) 23) '(= x -60) 4)
       (pick-prob
        (problem '(= (- 5 (* 3 x)) 56) '(= x -17) 4)
        (problem '(= (- 6 (* 3 x)) 39) '(= x -11) 4))
       (pick-prob
        (problem '(= (- 11 (* 1/9 x)) -45) '(= x 504) 4)
        (problem '(= (- 11 (* 9 x)) -43) '(= x 6) 4))
       (problem '(= 22 (+ 4 (* 5 x))) '(= x 18/5) 4)
       (problem '(= (- 15 x) 41) '(= x -26) 2)
       (problem '(= 13 (- 17 x)) '(= x 4) 2)
       (problem '(= (+ (* 3/4 x) 12) 36) '(= x 32) 4)
       ;; 26 skipped. decimals.
       (problem '(= (+ -8 (* 3 x)) 6) '(= x 14/3) 4)
       (problem '(= (- -2/5 x) 1/5) '(= x -3/5) 2))))
    (ch3-s1
     (values
      '(:no-canonicalize)
      '(evaluate combine collect reorder distribute undo-minus)
      (list
       (problem '(+ (* 5 (+ x 2)) 9) '((+ 19 (* 5 x))) 4 t))
      (list
       (problem '(+ (* 3 (+ x 5)) 11) '((+ 26 (* 3 x))) 4)
       (problem '(+ (* 7 (+ 6 x)) 8) '((+ (* 7 x) 50)) 4)
       (problem '(+ (* 2 (+ (* 3 x) 8)) 40) '((+ (* 6 x) 56)) 6)
       (problem '(- (* 9 (+ 4 (* 8 x))) 6) '((+ (* 72 x) 30)) 6)
       (problem '(+ 100 (* 6 (+ (* 4 x) 9))) '((+ (* 24 x) 154)) 6))))
    (ch3-s2 
     (values
      '(:no-canonicalize)
      '(evaluate combine collect reorder distribute undo-minus)
      (list
       (problem '(- 9 (* 2 (- x 4))) '((+ (* -2 x) 17)) 5 t))
      (list
       ;; 3-2
       (problem '(* 3 (+ x 7)) '((+ (* 3 x) 21)) 2)
       (problem '(* 7 (- (* 3 x) 8)) '((- (* 21 x) 56)) 4)
       ;; missing 5
       (problem '(* -5 (- 6 x)) '((- -30 (* -5 x))) 2)
       ;; missing 9
       ;;      (problem '(* x (+ x 9)) '((+ (* x x) (* x 9))))
       (problem '(* 6 (+ 9 (* 7 x))) '((+ 54 (* 42 x))) 4)
       ;; missing 15
       (problem '(* 28 (+ (* 1/4 x) 1/7)) '((+ (* 7 x) 4)) 4)
       (problem '(* 2/3 (+ (* 4 x) 3/4)) '((+ (* 8/3 x) 1/2)) 4)
       (problem '(- 8 (* 5 (- 4 (* 6 x)))) '((+ -12 (* 30 x))) 7)
       (problem '(- 11 (* 1 (+ (* 5 x) 3))) '((+ 8 (* -5 x))) 7))))
    (ch3-s4
     (values
      '()
      '(evaluate combine collect reorder distribute undo-minus)
      (list
       ;; we need a sample problem of each combine type
       (problem '(- (* 7 x) (* 5 x)) '((* 2 x)) 2 t)
       ;; This is perhaps Not Good. I made these examples up.
       (problem '(+ (* 5 x) (- 6 (* 2 x))) '((+ (* 3 x) 6)) 2 t))
      (list
       (problem '(+ (* 5 x) (* 9 x)) '((* 14 x)) 2)
       (problem '(- (* 8 x) (* 3 x)) '((* 5 x)) 2)
       (problem '(- (* 9 x) (* 8 x)) '((* 1 x)) 2)
       (problem '(+ (+ (* 3 x) (* 7 x)) 8) '((+ (* 10 x) 8)) 2)
       (problem '(+ (+ (* 5 x) 9) (* 7 x)) '((+ (* 12 x) 9)) 2)
       (problem '(- (- (* 3 x) 2) (* 8 x)) '((- (* -5 x) 2)) 2)
       (problem '(+ (- -4 (* 5 x)) (* 11 x)) '((+ -4 (* 6 x))) 2)
       (problem '(- (+ (* -6 x) 13) (* 7 x)) '((+ 13 (* -13 x))) 2)
       ;; This one might be too hairy
       ;;(problem '(+ (- (* 45/10 x) (* 71/10 x)) 116/10) '((+ 116/10 (* 116/10 x))) 1)
       ;; skip to #41
       ;; this uses reorder
       (problem '(+ (* 5 x) (* 3 (- x 2))) '((- (* 8 x) 6)) 4)
       (problem '(- 7 (* 4 (+ (* 2 x) 5))) '((- -13 (* 8 x))) 7)
       (problem '(- (* 2 (+ (* 3 x) 4)) (* 5 x)) '((+ (* 1 x) 8)) 6))))
    (ch3-s5
     (values
      '()
      '(evaluate combine collect reorder distribute canonicalize undo-minus)
      (list
       (problem '(+ (* 2 x) x) '((* 3 x)) 3 t)
       (problem '(- (+ (* 5 x) 3) (- 6 (* 2 x))) '((+ (* 7 x) -3)) 5 t))
      (list
       (problem '(+ (* -8 (- (* 9 x) 7)) (- (* 72 x) 57)) '(-1) 10)
       (problem '(- (* 6 x) x) '((* 5 x)) 3)
       (problem '(+ (* 4 (+ (* 3 x) 2)) (* 6 (+ (* 5 x) 8))) '((+ (* 42 x) 56)) 12)
       (problem '(+ (* 2 (- x 5)) (* 5 (- 4 (* 2 x)))) '((+ (* -8 x) 10)) 10)
       (problem '(- (* 6 x) (- x 1)) '((+ (* 5 x) 1)) 4)
       (problem '(- (* 8 (+ (* 2 x) 3)) (* 4 (+ (* 3 x) 6))) '((+ (* 4 x) 0)) 13)
       (problem '(- (* 3 (+ 6 (* 5 x))) (* 7 (- (* 2 x) 1))) '((+ (* 1 x) 25)) 13)
       (problem '(- (* -6 (- x 5)) (* 2 (- 7 (* 3 x)))) '(16) 13)
       (problem '(- (* 8 (+ x 4)) (- x 5)) '((+ (* 7 x) 37)) 8)
       ;; skip 63
       ;; this prolly won't work
       ;;(problem '(+ (* 3 x) (* (x (- 5 (* 2 x)))) '() 1)
       (problem '(- (* 1/4 (+ (* 12 x) 20)) (* 1/5 (- 30 (* 15 x)))) '((+ (* 6 x) -1)) 13)
       ;; skip 69
       )))
    (ch4-s1 
     (values 
      '()
      '(evaluate invert unwind combine collect reorder distribute canonicalize undo-minus)
      (list
       (problem '(= (+ (* 2 x) (* 6 x)) 56) '(= x 7) 4 t))
      (list
       (problem '(= (- (* 7 x) (* 4 x)) 15) '(= x 5) 4)
       (problem '(= (+ (+ (* 3 x) (* 4 x)) 8) 22) '(= x 2) 6)
       (problem '(= (+ (- (* 10 x) (* 7 x)) 18) 6) '(= x -4) 6)
       (problem '(= (+ (- (* 5 x) (* 7 x)) 21) 27) '(= x -3) 6)
       (problem '(= (+ (+ (* 7 x) 3) (* 3 x)) 63) '(= x 6) 6)
       (problem '(= (- (- (* 9 x) 14) (* 5 x)) -10) '(= x 1) 6)
       (problem '(= (- (* 10 x) x) 90) '(= x 10) 5)
       (problem '(= (- (- (* 7 x) 4) (* 6 x)) 5) '(= x 9) 6)
       (problem '(= (+ (+ (+ (* 2 x) (* 3 x)) 4) (* 5 x)) 34) '(= x 3) 8)
       (problem '(= (+ (+ (+ (* 3 x) 4) (* 5 x)) (* 6 x)) 32) '(= x 2) 8))))
    (ch4-s2
     (values
      '()
      '(evaluate invert unwind combine collect reorder distribute canonicalize undo-minus)
      (list 
       (problem '(= (+ (* 6 (- x 2)) (* 5 x)) 43) '(= x 5) 8 t))
      (list
       (problem '(= (+ (* 7 x) (* 3 x)) 40) '(= x 4) 4)
       (pick-prob
        (problem '(= (- (* 4 x) (* 11 x)) 105) '(= x -15) 4)
        (problem '(= (- (* 4 x) (* 11 x)) 49) '(= x -7) 4))
       (problem '(= (+ (+ (* 16 x) 3) (* 4 x)) 103) '(= x 5) 6)
       (problem '(= (- (- (* 7 x) 8) (* 4 x)) 25) '(= x 11) 6)
       (problem '(= (- (+ (* -5 x) 17) (* 8 x)) 56) '(= x -39/13) 6)
       (problem '(= 58 (+ (- 6 (* 14 x)) (* 12 x))) '(= x -26) 6)
       (problem '(= (- (+ (* 8 x) 5) (* 7 x)) 22) '(= x 17) 6)
       (problem '(= (- (+ (* 20 x) 84) (* 8 x)) 0) '(= x -7) 6)
       (problem '(= (+ (+ x 21) (* 4 x)) -33) '(= x -54/5) 7)
       (problem '(= (+ (+ (- (* 2 x) (* 5 x)) 13) (* 9 x)) 67) '(= x 9) 8)
       (problem '(= (- (* 5 (+ x 3)) (* 2 x)) -21) '(= x -12) 8)
       (problem '(= (+ (* 2 (- (* 3 x) 7)) (* 4 x)) 26) '(= x 4) 10)
       (problem '(= (+ (* 5 x) (* 3 (+ x 4))) 28) '(= x 2) 8)
       (problem '(= (- (* 7 x) (* 4 (- 2 (* 3 x)))) -27) '(= x -1) 11)
       (problem '(= (- (* 7 x) (+ 5 (* 6 x))) 4) '(= x 9) 7)
       (problem '(= (+ (- 12 (* 7 (- x 4))) x) -2) '(= x 7) 12)
       (problem '(= (- (* 2 (+ x 3)) (* 5 (- x 1))) 32) '(= x -7) 13)
       (problem '(= 55 (+ (* 3 (- (* 2 x) 1)) (* 2 (+ x 5)))) '(= x 6) 14)
       (problem '(= 0 (+ (* 4 (- 6 x)) (* 7 x))) '(= x -8) 8)
       (problem '(= (- (* 3 (+ x 2)) (- x 1)) 17) '(= x 5) 12))))
    (ch4-s3
     (values
      '()
      '(evaluate invert unwind combine collect reorder distribute canonicalize undo-minus subtract)
      (list
       (problem '(= (+ (* 3 x) 55) (* 8 x)) '(= x 11) 7 t))
      (list
       (problem '(= (+ (* 5 x) 27) (* 2 x)) '(= x -9) 7)
       (problem '(= (- (* 6 x) 28) (* 8 x)) '(= x -14) 7)
       (problem '(= (* 9 x) (- (* 4 x) 65)) '(= x -13) 8)
       (problem '(= (- 55 (* 3 x)) (* 8 x)) '(= x 5) 7)
       (problem '(= (* 5 x) (- 48 x)) '(= x 8) 9)
       (problem '(= (- (* 10 x) 51) (* 7 x)) '(= x 17) 7)
       (problem '(= (+ (* 29 x) 56) (* 27 x)) '(= x -28) 7)
       (problem '(= (* -6 x) (- x 70)) '(= x 10) 9)
       (problem '(= (* 9 x) (+ 34 (* 8 x))) '(= x 34) 8)
       (problem '(= (* 7 x) (- -16 (* 9 x))) '(= x -1) 8)
       (problem '(= (+ (* 5 x) 8) (+ (* 7 x) 8)) '(= x 0) 10)
       (problem '(= (- (* 2 x) 5) (+ (* 3 x) 4)) '(= x -9) 10)
       (problem '(= (- (+ (* 6 x) 7) (* 2 x)) (- (+ 3 (* 2 x)) 9)) '(= x -13/2) 14)
       (problem '(= (* 4 (+ x 3)) (* 6 x)) '(= x 6) 9)
       (pick-prob
        (problem '(= (* 5 (- 9 x)) (* 4 (+ x 18))) '(= x -3) 14)
        (problem '(= (* 3 (- 9 x)) (* 2 (+ x -4))) '(= x 7) 12))
       (problem '(= (+ (* 3 x) (* 3 (- 1 x))) (- x 17)) '(= x 20) 11)
       (problem '(= (* 2 (- 1 (* 3 (+ x 2)))) (- x 1)) '(= x -9/7) 22))))))

(defun gen-follow-instructions (simplify-p)
  (format nil "Follow the instructions ~a the problem to ~a."
          (if (or (tutor2-p)
                  *simplified-ui*)
              "below" "above")
          (if simplify-p
              "simplify it"
            "solve it")))

(defstruct section-instruction
  title text simplify-p)

(defun gen-instructions-tutor2 (section)
  (case section
    (ch1-s1
     (list
      (make-docs "Evaluate an Expression"
                 "To evaluate an expression find an operation with all numbers,
evaluate the operation, and repeat until there are no more
operations."
                 (gen-follow-instructions nil))))
    (ch1-s2
     (list
      (make-docs "Variables"
                 "A variable is a letter that represents a number. To
evaluate an expression first substitute the variable with its
number and then go on."
                 (gen-follow-instructions nil))))
    (ch1-s7
     (list
      (make-docs "Solving Equations"
                 "An equation is a sentence which says an expression with X is
equal to a number. To solve an equation, find the value
of X that would make this true. To do this the equation must be 
transformed so that X is alone on one side of the equal sign and a 
number is on the other side. You can do this by unwinding the operations on the side of the 
equation that contains X."
                 (gen-follow-instructions nil))))
    (ch2-s1
     (list
      (make-docs "Adding Negative Numbers"
                 "The absolute value of a number is the positive version of the number."
                 "If the two numbers have opposite signs, subtract the absolute values. The answer has the sign of the term with the greater absolute value."
                 "If the two numbers have the same sign, add the absolute values. Use the sign of the two numbers for the answer. Positive plus positive is positive. Negative plus negative is negative."
                 (gen-follow-instructions nil))))
    (ch2-s3
     (list
      (make-docs "Subtracting Signed Numbers"
                 "Change subtracting to adding the opposite. Then follow the rules for adding signed numbers."
                 (gen-follow-instructions nil))))
    (ch2-s4
     (list
      (make-docs "Multiplying Signed Numbers"
                 "Negative times negative is positive."
                 "Positive times negative is negative."
                 "Negative times positive is negative."
                 (gen-follow-instructions nil))))
    (ch2-s5
     (list
      (make-docs "Division Of Signed Numbers"
                 "Negative divided by negative is positive."
                 "Positive divided by negative is negative."
                 "Negative divided by positive is negative."
                 (gen-follow-instructions nil))))
    (ch2-s6
     (list
      (make-docs "Collecting Like Terms"
                 "When one of the expressions in an addition or subtraction is itself 
an addition or subtraction, you can combine the two operations. You 
can likewise combine two multiplications."
                 (gen-follow-instructions t))))
    (ch2-s7
     (list
      (make-docs "Equations That Need Two Transformations"
                 "The unwind transformation can be repeated."
                 (gen-follow-instructions nil))
      (make-docs "Unwind Special Case"
                 "There is a special case when the operator is - and the variable is on the right."
                 (gen-follow-instructions nil))))
    (ch3-s1
     (list
      (make-docs "The Distribute Transformation"
                 "If an expression has parenthesis in it, the operation inside the parenthesis should be done first."
                 "But, if the operation inside is addition and the operation outside is multiplication, you can multiply each number inside the parenthesis by the number outside the parenthesis and then add the answers."
                 (gen-follow-instructions t))))
    (ch3-s2
     (list
      (make-docs "Distributing over subtraction"
                 "You can also distribute when the operation inside the parenthesis is subtraction instead of addition."
                 (gen-follow-instructions t))))
    (ch3-s4
     (list
      (make-docs "Like Terms"
                 "The numeric coefficient of a term is the number that is multiplied by the variable. When two variables with numeric coefficients are added or subtracted you can use the Collect transformation to combine the numeric coefficients."
                 (gen-follow-instructions t))
      (make-docs "Like Terms"
                 "Collect works as long as all the operations between the variables are subtraction or addition."
                 (gen-follow-instructions t))))
    (ch3-s5
     (list
      (make-docs "Like Terms"
                 "TODO"
                 (gen-follow-instructions t))
      (make-docs "Like Terms"
                 "TODO"
                 (gen-follow-instructions t))))
    (ch4-s1
     (list
      (make-docs "Equations With Like Terms"
                 ;;"The Collect transformation can be used not only to collect numbers. It can be used to collect variables."
                 (gen-follow-instructions nil))
;;       (make-docs "Equations With Like Terms"
;;                  "Collect works when there is an extra + or - operation between the variables, too."
;;                  "Follow the instructions above the problem to simplify it as much as possible.")
      ))
    (ch4-s2
     (list
      (make-docs "Harder Equations"
                 (gen-follow-instructions nil))))
    (ch4-s3
     (list
      (make-docs "The Subtract Transformation"
                 "Sometimes the variable appears on both sides of the equation. To transform this into an equation one can solve, subtract the right side from the left and say the difference is 0."
                 (gen-follow-instructions nil))))))

(defparameter *tutor1-section-instruction*
  `((ch1-s1 
     ,(make-section-instruction
       :title "Evaluate a Diagram" 
       :text "Evaluating a diagram involves performing all the operations in the diagram to get a final result."
       :simplify-p nil))
    (ch1-s2
     ,(make-section-instruction 
       :title "Input Boxes" 
       :text "Some diagrams to be evaluated have input boxes with just a number that feed into the rest of the diagram."
       :simplify-p nil))
    (ch1-s7
     ,(make-section-instruction
       :title "Finding Input Values" 
       :text "Sometimes a diagram contains an input box without a number but at the bottom of the diagram there is a box with the value of the diagram. Then you must invert the diagram to find the value of the input box."
       :simplify-p nil))
    (ch2-s6
     ,(make-section-instruction
       :title "The Combine Operator" 
       :text "One can collapse two boxes with + or - into a single box and preserve the value of the diagram. One can do the same thing with two boxes with * or /."
       :simplify-p t))
    (ch2-s7
     ,(make-section-instruction
       :title "More On Invert" 
       :text "The invert operation can be repeated."
       :simplify-p nil)
     ,(make-section-instruction
       :title "More On Invert" 
       :text "There is a special case when the operator is - and the little black box is on the right."
       :simplify-p nil))
    (ch3-s1
     ,(make-section-instruction
       :title "The Reorder Operator" 
       :text "One can transform a diagram so that a + or - box above a * box moves to below the * box without changing the value of the diagram."
       :simplify-p t))
    (ch3-s2
     ,(make-section-instruction
       :title "Reorder And Subtraction" 
       :text "You can also reorder if a - box feeds into a * box."
       :simplify-p t)
     ,(make-section-instruction
       :title "Reorder And Subtraction" 
       :text "Before you combine boxes you must switch the operator and the signs in the box above if the little box is subtracted in the box below."
       :simplify-p t))
    (ch3-s4
     ,(make-section-instruction
       :title "More On Input Boxes" 
       :text "Sometimes an input box will have more than one arrow connecting it to other boxes. If the boxes connected to it are * boxes, and they are connected to a + or - box, then Combine can be used to simplify the diagram."
       :simplify-p t)
     ,(make-section-instruction
       :title "More On Input Boxes" 
       :text "Combine works when there is an extra + or - box, too."
       :simplify-p t))
    (ch3-s5
     ,(make-section-instruction
       :title "More On Input Boxes Part II"
       :text "If an input box is connected to a + or - box then you need to do an extra step before you can use combine."
       :simplify-p t)
     ,(make-section-instruction
       :title "More On Input Boxes Part II"
       :text "When two + or - boxes are connected via a + box then they can be combined. If they are connected via a - box, the connecting - must be \"undone\" first."
       :simplify-p t))
    (ch4-s1 ,(make-section-instruction :title "Find The Value Of The Input Box" :text "Use the knowledge you have acquired from the previous sections to find the value of the input box." :simplify-p nil))
    (ch4-s2 ,(make-section-instruction :title "Harder Diagrams" :text "Use the knowledge you have acquired from the previous sections to find the value of the input box." :simplify-p nil))
    (ch4-s3 
     ,(make-section-instruction
       :title "The Subtract Operation" 
       :text "When two boxes feed into the same result box at the bottom this means they are equal. One can transform this into a diagram we can operate on by subtracting the right box from the left and saying the difference is 0."
       :simplify-p nil))))

(defun gen-instructions-tutor1 (section)
  (let ((s (assoc section *tutor1-section-instruction*))
        ;; (angela-override (and *angela-experiment*
;;                               (assoc section *angela-instructions*)))
        )
    (if s
        ;; don't add the final line because the button often shows up
        ;; over top of it.
        (loop for i in (cdr s)
           collect (if *no-solve-instructions*
                       '((:text ""))
                       (if *force-read-instructions*
                           (make-docs (section-instruction-title i)
                                      (section-instruction-text i))
                           (make-docs (section-instruction-title i)
                                      (section-instruction-text i)
                                      (gen-follow-instructions (section-instruction-simplify-p i))))))
        (list (make-docs "no docs")))))


(defun get-section (section &optional tutor (problem-gen-fn *problem-generation-function*))
  ;; simplified-ui doesn't have a first mistake button. so jump
  ;; through some hoops to make sure it doesn't appear in that case.
  (let ((solve (apply #'make-docs (append (list "Solve the following problems"
                                                "If you get stuck, click the Hint button for a hint.")
                                          (if *simplified-ui* nil (list "If you make a mistake but can't find it, click the First Mistake button to jump back to it.")))))
        (simplify (apply #'make-docs (append (list (format nil "Simplify the following problems~%as much as possible")
                                                   "If you get stuck, click the Hint button for a hint.")
                                             (if *simplified-ui* nil (list "If you make a mistake but can't find it, click the First Mistake button to jump back to it."))))))
    (let ((instr (if tutor
                         (gen-instructions-tutor1 section)
                         (gen-instructions-tutor2 section))))
      (multiple-value-bind (hint-exclusions valids p1 p2) (funcall problem-gen-fn section)
        (append
         `((operations ,@valids))
         (mapcar (lambda (i p)
                   (make-problems i hint-exclusions (if (listp p)
                                                   p
                                                 (list p)))) instr p1)
         (list
          (make-problems
           ;; do they solve or simplify them?
           (if *sticky-instructions*
               (list* (caar instr)
                      (apply 'append (mapcar 'cdr instr)))
               ;; this is mega gross. we alread specify whats simplifyable.
             (cond (*no-solve-instructions*
                    '((:text "")))
                   ((find section '(ch2-s6 ch3-s1 ch3-s2 ch3-s4 ch3-s5))
                    simplify)
                   (t solve)))
           hint-exclusions
           p2)))))))

;; For backwards compatibility, generate the chN-sM functions
(macrolet ((defsection (section)
             `(defun ,section ()
                (get-section ',section (tutor1-p)))))
  (defsection ch1-s1)
  (defsection ch1-s2)
  (defsection ch1-s7)
  (defsection ch2-s1)
  (defsection ch2-s3)
  (defsection ch2-s4)
  (defsection ch2-s5)
  (defsection ch2-s6)
  (defsection ch2-s7)
  (defsection ch3-s1)
  (defsection ch3-s2)
  (defsection ch3-s4)
  (defsection ch3-s5)
  (defsection ch4-s1)
  (defsection ch4-s2)
  (defsection ch4-s3))

(defun do-section (ch section &optional (subject 1))
  "Run the tutor on chapter CH, section SECTION."
  (check-type ch number)
  (check-type section number)
  (let ((sym (intern (format nil "CH~a-S~a" ch section) "TUTOR")))
    (tutor subject (nconc
                    (funcall sym)
                    '((end))))))

(defun do-problem (eqn &optional (answer '(1)) (subject 1))
  (setf *valid-operations* (all-operations nil))
  (tutor subject
         (list
          (make-problems 
           '((:title "Problem")
             (:text "Solve the problem."))
           '()
           (list
            (problem eqn answer 1)))
          '(end))))

(defun build-session-sections (sections &optional (section-fn #'funcall))
  (apply #'nconc (loop for s on (mapcar section-fn sections)
                       collect (list (car s)
                                     (if (cdr s)
                                         '((break))
                                         '((end)))))))

(defun generate-user-hints-for (section)
  (list '(eval
          ;;(push :user *hint-types*)
          (setf *force-read-instructions* nil))
        (make-problems `((:title "Your Turn!")
                         (:input "Click here to enter your instructions"))
                       nil
                       (ecase section
                         (ch1-s1 (list (problem '(* (+ 5 7) 8) '(96) 1 nil :user-hint-for-evaluate)))
                         (ch1-s2 (list (problem '(+ 5 (4)) '(9) 1 nil :user-hint-for-evaluate)))
                         (ch1-s7 (list (problem '(= (+ x 3) 8) '(= x 5) 2)))
                         (ch2-s6 (list (problem '(+ 3 (+ (* 2 x) 7)) '((+ (* 2 x) 10)) 2)))
                         (ch2-s7 (list (problem '(= (+ (* 2 x) 3) 19) '(= x 8) 4)
                                       (problem '(= (- 10 x) 2) '(= x 8) 2)))
                         (ch3-s1 (list (problem '(+ (* 5 (+ x 2)) 9) '((+ 19 (* 5 x))) 4)))
                         (ch3-s2 (list (problem '(- 9 (* 2 (- x 4))) '((+ (* -2 x) 17)) 4)))
                         (ch3-s4 (list (problem '(- (* 7 x) (* 5 x)) '((* 2 x)) 2)
                                       (problem '(+ (* 5 x) (- 6 (* 2 x))) '((+ (* 3 x) 6)) 2)))
                         (ch3-s5 (list (problem '(+ (* 2 x) x) '((* 3 x)) 2)
                                       (problem '(- (+ (* 5 x) 3) (- 6 (* 2 x))) '((+ (* 7 x) -3)) 2)))
                         (ch4-s1 (list (problem '(= (+ (* 2 x) (* 6 x)) 56) '(= x 7) 4)))
                         (ch4-s2 (list (problem '(= (+ (* 6 (- x 2)) (* 5 x)) 43) '(= x 5) 8)))
                         (ch4-s3 (list (problem '(= (+ (* 3 x) 55) (* 8 x)) '(= x 11) 7)))))
        '(eval
          (setf ;;*hint-types* (remove :user *hint-types*)
                *force-read-instructions* t))))

(defun build-angela-sections (sections &optional (instr-gen t) (end '((end))))
  (apply #'nconc (loop for s on (mapcar #'funcall sections)
                    for i in sections
                    collect (if instr-gen 
                                (list (car s)
                                      '((break "Now it's your turn to supply the hints!"))
                                      (generate-user-hints-for i)
                                      (if (cdr s)
                                          '((break))
                                          end))
                                ;; skip the instruction generation
                                (list (car s)
                                      (if (cdr s)
                                          '((break))
                                          end))))))

(defun tutor1-session-data (n)
  (ecase n
    (1 (values '(ch1-s1 ch1-s2 ch1-s7 ch2-s6 ch2-s7)
               "14.98"))
    (2 (values '(ch3-s1 ch3-s2 ch3-s4 ch4-s1)
               "15.82"))
    (3 (values '(ch4-s2 ch4-s3)
               "22.54"))
    (10 (values '(ch1-s1 ch1-s2 ch1-s7 ch2-s1 ch2-s3 ch2-s4 ch2-s5 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch4-s1 ch4-s2 ch4-s3)
                "0.00"))))

(defun tutor2-session-data (n)
   (ecase n
     (1 (values '(ch1-s1 ch1-s2 ch1-s7 ch2-s1 ch2-s3 ch2-s4 ch2-s5 ch2-s6 ch2-s7)
                "23.31"))
     (2 (values '(ch3-s1 ch3-s2 ch3-s4 ch4-s1)
                "15.82"))
     (3 (values '(ch4-s2 ch4-s3)
                "22.54"))
     (10 (values '(ch1-s1 ch1-s2 ch1-s7 ch2-s1 ch2-s3 ch2-s4 ch2-s5 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch4-s1 ch4-s2 ch4-s3)
                 "0.00"))))

(defun tutor1-session (n id)
  (multiple-value-bind (sections money) (tutor1-session-data n)
    (tutor id
           (append
            `((text ((:title ,(format nil "Algebra Tutor Session ~a" n))
                     (:text "In this experiment you will solve a number of problems given some instruction.")
                     (:text "In each problem there will be boxes
containing arithmetic expressions like 3+4. Some boxes will be
connected to other boxes with arrows. Your goal will be to clean
up the problems by reducing the numbers of boxes as much as
possible using rules explained in each section.")
                     (:text ,(format nil "Session ~a problems are broken up into ~a sections. The
first and sometimes the second problem in each section is a practice problem. The tutor
will guide you through each step. After the practice problem you
will be given a number of real problems to solve." n (length sections)))
                     (:text "Instructions will appear in the top-left corner of the tutor window. Read them carefully."))
                    ((:title "The Incentive System")
                     (:text ,(format nil "Each problem is worth a certain amount of money based on
how many operations are needed to solve it. If you complete every problem perfectly you
will make $~a. In the lower right corner of the tutor window
you can see how much money you've made so far." money))
                     (:text "Each operation is worth 7 cents. A mistake deducts some amount from the problem's total worth.")
                     (:text "If you choose an operation that can't be performed with
the boxes you've selected, you will lose 3 cents for the problem.")
                     (:text "If you select the right boxes and choose the right
operation, but fill in the wrong values you will lose 7 cents.")
                     (:text "You can never make less than $0 on a problem.")
                     (:text "Practice problems are not rewarded."))
                    ((:title "Hints and Mistakes")
                     (:text "If you get stuck or are unsure of what to do next, you
can ask the tutor for a hint by clicking the 'Hint' button. A hint costs 4 cents.")
                     (:text "The tutor will not let you advance to the next problem
until you have solved the current one correctly. If you make a
mistake you can use the 'First Mistake' button to jump back to
the diagram state after your mistake. The diagram on the left side of the tutor is the diagram state before your mistake was made. Use it
with the diagram on the right to identify your error. Once you have, click the '<-' button to go
back, then redo the operation correctly."))
                    ((:title "The Calculator")
                     (:text "You have a calculator at your disposal to help you with the arithmetic. Take a moment to familiarize yourself with its functionality. Note that the  button is used to create fractions."))
                    ((:title "Questions?")
                     (:text "If you have any questions, ask your experimenter now. If not click 'Begin' to start the tutor."))))
            (apply #'append (build-session-sections sections))))))

(defun tutor2-session (n id)
  (multiple-value-bind (sections money) (tutor2-session-data n)
    (tutor id
           (append
            `((text ((:title ,(format nil "Algebra Tutor Session ~a" n))
                     (:text "In this experiment you will solve a number of problems given some instruction.")
                     (:text "Each problem will have arithmetic expressions in it such as 3+4. To solve the problem you will use different
operations to transform the problem into the answer.")
                     (:text ,(format nil "Session ~a problems are broken up into ~a sections. The
first and sometimes the second problem in each section is a practice problem. The tutor
will guide you through each step. After the practice problem you
will be given a number of real problems to solve." n (length sections)))
                     (:text "Instructions will appear in the top-left corner of the tutor window. Read them carefully."))
                    ((:title "The Incentive System")
                     (:text ,(format nil "Each problem is worth a certain amount of money based on
how many operations are needed to solve it. If you complete every problem perfectly you
will make $~a. In the lower right corner of the tutor window
you can see how much money you've made so far." money))
                     (:text ,(format nil "Each operation is worth ~d cents. Mistakes cost you." +operation-points+))
                     (:text ,(format nil "If you choose an operation that can't be performed with
the operators you've selected, you will lose ~d cents for the problem." +bad-operation-points+))
                     (:text ,(format nil "If you select the right operators and choose the right
operation, but fill in the wrong values you will lose ~d cents." +bad-fill-in+))
                     (:text "You can never make less than $0 on a problem.")
                     (:text "Practice problems are not rewarded."))
                    ((:title "Hints and Mistakes")
                     (:text ,(format nil "If you get stuck or are unsure of what to do next, you
can ask the tutor for a hint by clicking the 'Hint' button. A hint costs ~d cents." +hint-points+))
                     (:text "The tutor will not let you advance to the next problem
until you have solved the current one correctly. If you make a
mistake you can use the 'First Mistake' button to jump back to
the problem state after your mistake. Use it
with the line above to identify your error. Once you have, click the '<-' button to go
back, then redo the operation correctly."))
                    ((:title "The Calculator")
                     (:text "You have a calculator at your disposal to help you with the arithmetic. Take a moment to familiarize yourself with its functionality. Note that the  button is used to create fractions."))
                    ((:title "Questions?")
                     (:text "If you have any questions, ask your experimenter now. If not click 'Begin' to start the tutor."))))
            (apply #'append (build-session-sections sections))))))

(defun angela-regular-intro (session sections)
  `((text ((:title ,(format nil "Algebra Tutor Session ~a" session))
           (:text "In this experiment you will solve a number of problems given some instruction.")
                     (:text "In each problem there will be boxes
containing arithmetic expressions like 3+4. Some boxes will be
connected to other boxes with arrows. Your goal will be to clean
up the problems by reducing the numbers of boxes as much as
possible using rules explained in each section.")
           (:text ,(format nil "Session ~a problems are broken up into ~a sections. The
first and sometimes the second problem in each section is a practice problem. The tutor
will guide you through each step. After the practice problem you
will be given a number of real problems to solve." session sections))
           (:text "Instructions will appear in the top-left corner of the tutor window. Read them carefully."))
          ((:title "The Reward System")
           (:text "Each problem is worth a certain amount of money based on how many operations are needed to solve it.")
           (:text ,(format nil "Each operation is worth ~d cents. Mistakes cost you." +operation-points+))
           (:text ,(format nil "If you choose an operation that can't be performed with
the boxes you've selected, you will lose ~d cents for the problem." +bad-operation-points+))
           (:text ,(format nil "If you select the right boxes and choose the right
operation, but fill in the wrong values you will lose ~d cents." +bad-fill-in+))
           (:text "You can never make less than $0 on a problem.")
           (:text "Practice problems are not rewarded."))
          ((:title "Hints and Mistakes")
           (:text ,(format nil "If you get stuck or are unsure of what to do next, you
can ask the tutor for a hint by clicking the 'Hint' button. A hint costs ~d cents." +hint-points+))
           (:text "The tutor will not let you advance to the next problem
until you have solved the current one correctly. If you make a
mistake you can use the 'First Mistake' button to jump back to
the problem state before you made any mistakes."))
          ((:title "The Calculator")
           (:text "You have a calculator at your disposal to help you with the arithmetic. Take a moment to familiarize yourself with its functionality. Note that the / button is used to create fractions."))
          ((:title "Questions?")
           (:text "If you get stuck while using the tutor and the hints don't help, you can use the Ask Experimenter button to summon the experimenter. When you click it, the tutor will freeze and you can call the experimenter into the room. Once finished, click the Resume Problem button to unfreeze the tutor.")
           (:text "If you have any questions about what you just read, ask your experimenter now. If not click 'Begin' to start the tutor.")))))

(defun angela-intro (session sections)
  `((text ((:title ,(format nil "Algebra Tutor Session ~a" session))
           (:text "In this experiment you will solve a number of problems given some instruction.")
                     (:text "In each problem there will be boxes
containing arithmetic expressions like 3+4. Some boxes will be
connected to other boxes with arrows. Your goal will be to clean
up the problems by reducing the numbers of boxes as much as
possible using rules explained in each section.")
           (:text ,(format nil "Session ~a problems are broken up into ~a sections. The
first and sometimes the second problem in each section is a practice problem. The tutor
will guide you through each step. After the practice problem you
will be given a number of real problems to solve." session sections))
           (:text "Instructions will appear in the top-left corner of the tutor window. Read them carefully."))
          ((:title "The Reward System")
           (:text "Each problem is worth a certain amount of money based on how many operations are needed to solve it.")
           (:text ,(format nil "Each operation is worth ~d cents. Mistakes cost you." +operation-points+))
           (:text ,(format nil "If you choose an operation that can't be performed with
the boxes you've selected, you will lose ~d cents for the problem." +bad-operation-points+))
           (:text ,(format nil "If you select the right boxes and choose the right
operation, but fill in the wrong values you will lose ~d cents." +bad-fill-in+))
           (:text "You can never make less than $0 on a problem.")
           (:text "Practice problems are not rewarded."))
          ((:title "Hints and Mistakes")
           (:text ,(format nil "If you get stuck or are unsure of what to do next, you
can ask the tutor for a hint by clicking the 'Hint' button. A hint costs ~d cents." +hint-points+))
           (:text "The tutor will not let you advance to the next problem
until you have solved the current one correctly. If you make a
mistake you can use the 'First Mistake' button to jump back to
the problem state before you made any mistakes."))
          ((:title "The Calculator")
           (:text "You have a calculator at your disposal to help you with the arithmetic. Take a moment to familiarize yourself with its functionality. Note that the / button is used to create fractions."))
          ((:title "Instruction Generation")
           (:text "At the end of each section you will be asked to write step by step instructions for the practice problems you did at the beginning. Imagine someone was given no instruction but what you provided and asked to solve the same problems you did. What would tell them?"))
          ((:title "Questions?")
           (:text "If you get stuck while using the tutor and the hints don't help, you can use the Ask Experimenter button to summon the experimenter. When you click it, the tutor will freeze and you can call the experimenter into the room. Once finished, click the Resume Problem button to unfreeze the tutor.")
           (:text "If you have any questions about what you just read, ask your experimenter now. If not click 'Begin' to start the tutor.")))))

;; Generate the sessions
(macrolet ((defsession (n)
             (with-gensyms (id)
               `(defun ,(intern (format nil "SESSION~a" n) "TUTOR") (,id)
                  (if (tutor1-p)
                      (tutor1-session ,n ,id)
                    (tutor2-session ,n ,id))))))
  (defsession 1)
  (defsession 2)
  (defsession 3)
  ;; session 10 contains all the sections
  (defsession 10))

(defun fmri-split-section (section)
  "Return an fsm list with the problems split into scan periods."
  ;; yuck. but hey, it works.
  (let ((splits (ecase section
                  (ch1-s7 (list 8 16))
                  (ch2-s6 (list 7 14))
                  (ch2-s7 (list 5 10))))
        (instr (if (tutor1-p)
                   (gen-instructions-tutor1 section)
                 (gen-instructions-tutor2 section))))
    (multiple-value-bind (hint-exclusions p1 p2) (gen-problems section)
      ;; create the practice problem(s)
      (append
       (mapcar (lambda (i p)
                 (make-problems i hint-exclusions (if (listp p)
                                                 p
                                               (list p)))) instr p1)
       (apply 'append (loop with blocks = (apply 'split-list p2 splits)
                            for block in blocks
                            for i from 1
                            collect (list (list 'wait (format nil "Get ready for block ~a of ~a~%Please wait for experimenter" i (length blocks)))
                                          (list 'crosshair *crosshair-before-problem-time*)
                                          (make-problems nil hint-exclusions block))))))))
    
(defun fmri-instructions ()
  `((text ((:title "Algebra Tutor")
           (:text "In this experiment you will solve a number of problems given some instruction.")
           (:text "Each problem will have arithmetic expressions in it such as 3+4. All answers are either whole numbers or fractions. There are no decimal answers.")
           (:title "Input Drill")
           (:text "The first part of this experiment will be getting comfortable with the input window. Note that the '' button is used to create fractions."))
          ((:title "The Tutor")
           (:text "The rest of the experiment will be solving algebra problems. Each section has a practice problem (some have 2 practice problems) at the beginning where the tutor guides you through.")
           (:text "When you perform an operation, the diagram will change and you will have to fill in the missing parts. The tutor will not let you move on until they are all correctly filled in.")
           (:text "To help you fill in the missing parts, the relevant portions of the diagram before it changed will be shown in the gray rectangle beside the diagram.")
           (:text "Answers will always be whole numbers or fractions. There are no decimal numbers. Fractions must be reduced. The tutor will accept 1/2 but not 2/4.")
           (:text "If you are unable to solve a problem you can click the Hint button which will tell you which boxes to select to perform an operation, which operation to perform if you have the right boxes selected, or what values to fill in depending on what stage in solving the problem you're at."))
          ((:title "Questions?")
           (:text "If you have any questions please ask your experimenter now or click Begin to start.")))))

(defun fmri-nback-instructions ()
  "These instructions appear at the beginning of session 2."
  `((text ((:title "Scan Session")
           (:text "This session is broken into 2 parts. The first part will not be scanned and consists of two sections. The second part will be scanned and consists of 3 sections.")
           (:text "Each section in the second part will be broken into scanning blocks of between 5 and 8 problems.")
           (:title "Extra Task")
           (:text "An extra task has been added to this session. It happens everytime you finish filling in the green boxes. The problem will disappear and letters will be displayed one after another. Whenever the last letter is the same as the current one, you must click the Match! button."))
          ((:title "Questions?")
           (:text "If you have any questions please ask your experimenter now or click Begin to start.")))))

(defun fmri-session1 (subject)
  "Out of the magnet. Shawn looking over their shoulder. No
n-back. The 'alternative' problem set is done, leaving 1-1 odd and even
for the structurals."
  (setup-tutor1)
  (fmri-enable)
  (setf *odd-problems* :alternative
        *0back-after-submit* nil
        *crosshair-before-problem* nil
        ;; tutor normally sets this but calc-practice needs it.
        *history-id* subject)
  (history-add 'fmri-session1 subject :alternative)
  (tutor subject (append (fmri-instructions) `((calc ,(lambda ()))) (apply 'append (build-session-sections '(ch1-s1))))))

(defun fmri-session2-part1 (subject &optional (odd-p (oddp subject)))
  "While the structural is taking place. 1-1 and 1-2 with n-back."
  (setup-tutor1)
  (fmri-enable)
  (setf *odd-problems* odd-p
        *crosshair-before-problem* nil
        *history-id* subject)
  (history-add 'fmri-session2-part1 subject odd-p)
  (tutor subject (append (fmri-nback-instructions) (apply #'append (build-session-sections '(ch1-s1 ch1-s2))))))

(defun fmri-session2-part2 (subject &optional (odd-p (oddp subject)))
  "In the magnet. The Real Deal."
  (setup-tutor1)
  (fmri-enable)
  (setf *odd-problems* odd-p
        *crosshair-before-problem* t
        *history-id* subject)
  (history-add 'fmri-session2-part2 subject odd-p)
  (let* ((states (append (fmri-split-section 'ch1-s7)
                         `((wait ,(format nil "1 of 3 sections finished!~%Please wait for experimenter")))
                         (fmri-split-section 'ch2-s6)
                         `((wait ,(format nil "2 of 3 sections finished!~%Please wait for experimenter")))
                         (fmri-split-section 'ch2-s7)
                         '((end))))
         (panel (fsm-panel states)))
    (set-view-position panel 0 20)
    (tutor subject states)))

(defun fmri-session3 (subject)
  "Outside of the magnet."
  (setup-tutor1)
  (fmri-enable)
  (setf *0back-after-submit* nil
        *crosshair-before-problem* nil
        *history-id* subject)
  (history-add 'fmri-session3 subject)
  (tutor subject (apply #'append (build-session-sections '(ch3-s1 ch3-s2 ch3-s4 ch4-s1)))))

(defun fmri-session4 (subject)
  "Outside of the magnet."
  (setup-tutor1)
  (fmri-enable)
  (setf *0back-after-submit* nil
        *crosshair-before-problem* nil
        *history-id* subject)
  (history-add 'fmri-session4 subject)
  (tutor subject (apply #'append (build-session-sections '(ch4-s2 ch4-s3)))))

(defun fmri-session4-type2 (subject)
  "Sometimes subjects can't finish ch4-s1 in time, so here's a session with ch4-s1 included."
  (setup-tutor1)
  (fmri-enable)
  (setf *0back-after-submit* nil
        *crosshair-before-problem* nil
        *history-id* subject)
  (history-add 'fmri-session4-type2 subject)
  (tutor subject (apply #'append (build-session-sections '(ch4-s1 ch4-s2 ch4-s3)))))

(defun fmri-session5-part1 (subject &optional (odd-p (oddp subject)))
  "during the structural."
  ;; If they did odd the first time, then don't do odd this time
  (fmri-session2-part1 subject (not odd-p)))

(defun fmri-session5-part2 (subject &optional (odd-p (oddp subject)))
  "in the magnet."
  ;; If they did odd the first time, then don't do odd this time
  (fmri-session2-part2 subject (not odd-p)))

;;; Algela's experiment sections

(defmacro def-angela-session (session &rest sections)
  (with-gensyms (type subject instr-gen)
    `(defun ,(intern (format nil "ANGELA-SESSION~d" session)) (&optional (,instr-gen t) (,type (progn (format t "1. No Hints~%2. Direct Demonstration~%3. Written Instruction~%4. Both~%") (read))) (,subject (progn (princ "Subject number: ") (read))))
       (check-type ,subject number)
       (check-type ,type number)
       (ecase ,type
         (1 (angela-experiment nil nil))
         (2 (angela-experiment t nil))
         (3 (angela-experiment nil t))
         (4 (angela-experiment t t)))
       (unless (and calc::*calc-window*
                    (window-shown-p calc::*calc-window*))
         (calc:calc))
       ;; store the condition in the log file. FIXME: We need to set
       ;; the history id to log the history.
       (setf *history-id* (format nil "~a-session~a" ,subject ,session))
       (tutor *history-id*
              (apply #'append (if ,instr-gen
                                  (angela-intro ,session ,(length sections))
                                  (angela-regular-intro ,session ,(length sections)))
                     (build-angela-sections ',sections ,instr-gen))
              (list :angela-condition ,type)))))

(def-angela-session 1 ch1-s1 ch1-s2 ch1-s7 ch2-s6 ch2-s7)
(def-angela-session 2 ch3-s1 ch3-s2 ch3-s4 ch3-s5)
(def-angela-session 3 ch4-s1 ch4-s2 ch4-s3)

;; (defun angela-generate-instructions ()
;;   "Do the practice problems for each section."
;;   ;;   (setf *hint-types* nil)
;;   (setf *sticky-instructions* nil
;;         *force-read-instructions* nil
;;         *ask-experimenter-button* t
;;         *hint-types* (list :user))
;;   (tutor 0 
;;          (list (make-problems `((:title "Your Turn!")
;;                                 (:input "Click here to write your instructions"))
;;                               nil
;;                               (list (problem '(* (+ 5 7) 8) '(96) 1 t)
;;                                     (problem '(+ 5 (4)) '(9) 1 t)
;;                                     (problem '(= (+ x 3) 8) '(= x 5) 2 t)
;;                                     (problem '(+ 3 (+ (* 2 x) 7)) '((+ (* 2 x) 10)) 2 t)
;;                                     (problem '(= (+ (* 2 x) 3) 19) '(= x 8) 4 t)
;;                                     (problem '(= (- 10 x) 2) '(= x 8) 2 t)
;;                                     (problem '(+ (* 5 (+ x 2)) 9) '((+ 19 (* 5 x))) 4 t)
;;                                     (problem '(- 9 (* 2 (- x 4))) '((+ (* -2 x) 17)) 4 t)
;;                                     (problem '(- (* 7 x) (* 5 x)) '((* 2 x)) 2 t)
;;                                     (problem '(+ (* 5 x) (- 6 (* 2 x))) '((+ (* 3 x) 6)) 2 t)
;;                                     (problem '(+ (* 2 x) x) '((* 3 x)) 2 t)
;;                                     (problem '(- (+ (* 5 x) 3) (- 6 (* 2 x))) '((+ (* 7 x) -3)) 2 t)
;;                                     (problem '(= (+ (* 3 x) 55) (* 8 x)) '(= x 11) 7 t)))
;;                '(wait "Thanks!"))))

(defun test-problem-type ()
  (tutor 1 `(,(make-problem-template '((:text "Easy combine"))
                                     '(= :n1 (:a (* :n2 x) (* :n3 x))) '(= x (:eval / :n1 (:eval :a :n2 :n3))) 3)
              ,(make-problem-template '((:text "Combine type 2!"))
                                      '(= :n1 (:a1 (* :n2 x) (:a2 (* :n3 x) :n4))) '(= x (:eval / (:eval - (:eval :a2 :n4) :n1) (:eval :a1 :n2 (:eval :a2 :n3)))) 3)
              ,(make-problem-template '((:text "canonicalize!"))
                                      '(:a1 (* :n2 x) (:a2 x :n4)) '((+ (:eval :a1 (:eval :a2 :n4)) (* (:eval :a1 :n2 1) x))) 3)
              )))

(defun do-all-sections ()
  (tutor 1 (apply #'append (angela-intro 1 7) (build-session-sections '(ch1-s1 ch1-s2 ch1-s7 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch4-s1 ch4-s2 ch4-s3)))))

(defun test-combine ()
  (setf *force-correct-fill-in* t)
  (tutor 1 (list (make-problems '((:text "go!")) nil
    (list
     (problem (prefixify '(5 + (7 + x))) '(1) 1 t)
     (problem (prefixify '(5 + (b + 7))) '(1) 1 t)
     (problem (prefixify '((7 + x) + 5)) '(1) 1 t)
     (problem (prefixify '((b + 7) + 5)) '(1) 1 t)
     (problem (prefixify '(5 + (7 - x))) '(1) 1 t)
     (problem (prefixify '((7 - x) + 5)) '(1) 1 t)
     (problem (prefixify '(5 + (b - 7))) '(1) 1 t)
     (problem (prefixify '((b - 7) + 5)) '(1) 1 t)
     (problem (prefixify '(5 - (7 + x))) '(1) 1 t)
     (problem (prefixify '(5 - (b + 7))) '(1) 1 t)
     (problem (prefixify '((7 + x) - 5)) '(1) 1 t)
     (problem (prefixify '((b + 7) - 5)) '(1) 1 t)
     (problem (prefixify '((7 - x) - 5)) '(1) 1 t)
     (problem (prefixify '((b - 7) - 5)) '(1) 1 t)
     (problem (prefixify '(5 * (7 * x))) '(1) 1 t)
     (problem (prefixify '(5 * (b * 7))) '(1) 1 t)
     (problem (prefixify '((7 * x) * 5)) '(1) 1 t)
     (problem (prefixify '((b * 7) * 5)) '(1) 1 t)
     (problem (prefixify '((7 * x) / 5)) '(1) 1 t)
     (problem (prefixify '((5 * x) + (7 * x))) '(1) 1 t)
     (problem (prefixify '((5 * x) - (7 * x))) '(1) 1 t)
     (problem (prefixify '((5 * x) + (9 + (7 * x)))) '(1) 1 t)
     (problem (prefixify '((5 * x) + (9 - (7 * x)))) '(1) 1 t)
     (problem (prefixify '((5 * x) + ((7 * x) + 9))) '(1) 1 t)
     (problem (prefixify '((5 * x) + ((7 * x) - 9))) '(1) 1 t)
     (problem (prefixify '((9 + (7 * x)) + (5 * x))) '(1) 1 t)
     (problem (prefixify '((9 - (7 * x)) + (5 * x))) '(1) 1 t)
     (problem (prefixify '((9 + (7 * x)) - (5 * x))) '(1) 1 t)
     (problem (prefixify '((9 - (7 * x)) - (5 * x))) '(1) 1 t)
     (problem (prefixify '(((7 * x) + 9) + (5 * x))) '(1) 1 t)
     (problem (prefixify '(((7 * x) + 9) - (5 * x))) '(1) 1 t)
     (problem (prefixify '(((7 * x) - 9) + (5 * x))) '(1) 1 t)
     (problem (prefixify '(((7 * x) - 9) - (5 * x))) '(1) 1 t))))))

(defun test-combine-type-3 ()
  (setf *force-correct-fill-in* t)
  (tutor 1 (list (make-problems '((:text "go!")) nil
    (list
     (problem (prefixify '(((3 * x) + 1) - ((5 * x) + 2))) '(1) 1 t)
     (problem (prefixify '((1 + (3 * x)) - ((5 * x) + 2))) '(1) 1 t)
     (problem (prefixify '(((3 * x) + 1) - (2 + (5 * x)))) '(1) 1 t)
     (problem (prefixify '((1 + (3 * x)) - (2 + (5 * x)))) '(1) 1 t)
     (problem (prefixify '(((3 * x) + 1) - ((5 * x) - 2))) '(1) 1 t)
     (problem (prefixify '((1 + (3 * x)) - ((5 * x) - 2))) '(1) 1 t)
     (problem (prefixify '(((3 * x) + 1) - (2 - (5 * x)))) '(1) 1 t)
     (problem (prefixify '((1 + (3 * x)) - (2 - (5 * x)))) '(1) 1 t)
     (problem (prefixify '(((3 * x) - 1) - ((5 * x) + 2))) '(1) 1 t)
     (problem (prefixify '(((3 * x) - 1) - (2 + (5 * x)))) '(1) 1 t)
     (problem (prefixify '((1 - (3 * x)) - ((5 * x) + 2))) '(1) 1 t)
     (problem (prefixify '((1 - (3 * x)) - (2 + (5 * x)))) '(1) 1 t)
     (problem (prefixify '(((3 * x) - 1) - ((5 * x) - 2))) '(1) 1 t)
     (problem (prefixify '(((3 * x) - 1) - (2 - (5 * x)))) '(1) 1 t)
     (problem (prefixify '((1 - (3 * x)) - ((5 * x) - 2))) '(1) 1 t)
     (problem (prefixify '((1 - (3 * x)) - (2 - (5 * x)))) '(1) 1 t))))))

;;; Angela's shortened tutor

(defun angela-short-intro (sections global-instruction-p direct-demonstration-p)
  `((text ((:title "Algebra Tutor")
           (:text "In this experiment you will solve a number of problems.")
           (:text "In each problem there will be boxes
containing arithmetic expressions like 3+4. Some boxes will be
connected to other boxes with arrows. Your goal will be to clean
up the problems by reducing the numbers of boxes as much as
possible.")
           (:text ,(format nil "There will be ~a sections." sections))
           ,@(when global-instruction-p
                `((:text "Instructions will appear in the top-left corner of the tutor window. Read them carefully."))))
          ((:title "The Reward System")
           (:text "Each problem is worth a certain amount of money based on how many operations are needed to solve it.")
           (:text ,(format nil "Each operation is worth ~d cents. Mistakes cost you." +operation-points+))
           (:text ,(format nil "If you choose an operation that can't be performed with
the boxes you've selected, you will lose ~d cents for the problem." +bad-operation-points+))
           (:text ,(format nil "If you select the right boxes and choose the right
operation, but fill in the wrong values you will lose ~d cents." +bad-fill-in+))
           (:text "You can never make less than $0 on a problem.")
           ,@(when direct-demonstration-p
                `((:text "Practice problems are not rewarded."))))
          ((:title "Hints and Mistakes")
           ,@(when direct-demonstration-p
                   `((:text ,(format nil "If you get stuck or are unsure of what to do next, you
can ask the tutor for a hint by clicking the 'Hint' button. A hint costs ~d cents." +hint-points+))))
           (:text "The tutor will not let you advance to the next problem
until you have solved the current one correctly. If you make a
mistake you can use the 'First Mistake' button to jump back to
the problem state before you made any mistakes."))
          ((:title "The Calculator")
           (:text "You have a calculator at your disposal to help you with the arithmetic. Take a moment to familiarize yourself with its functionality. Note that the / button is used to create fractions."))
          ((:title "Questions?")
           ,(if direct-demonstration-p
                `(:text "If you get stuck while using the tutor and the hints don't help, you can use the Ask Experimenter button to summon the experimenter. When you click it, the tutor will freeze and you can call the experimenter into the room. Once finished, click the Resume Problem button to unfreeze the tutor.")
                `(:text "If you get stuck while using the tutor, you can use the Ask Experimenter button to summon the experimenter. When you click it, the tutor will freeze and you can call the experimenter into the room. Once finished, click the Resume Problem button to unfreeze the tutor."))
           (:text "If you have any questions about what you just read, ask your experimenter now. If not click 'Begin' to start the tutor.")))))

(defun get-short-section (section)
  (get-section section t 'gen-short-problems))

(defun gen-short-problems (section)
  (ecase section
    (ch1-s1
     (values
      '() '(evaluate)
      (list
       (problem '(* (+ 5 7) 8) '(96) 2 t))
      (list
       (problem '(- 9 (* 4 2)) '(1) 2)
       (problem '(* 5 (/ (- 19 7) (+ 5 1))) '(10) 4))))
    (ch1-s2
     (values
      '() '(evaluate)
      (list
       (problem '(+ 5 (4)) '(9) 2 t))
      (list
       (problem '(+ (93) 5) '(98) 2)
       (problem '(/ (+ (0) 7) 3) '(7/3) 3))))
    (ch1-s7
     (values
      '() '(evaluate invert)
      (list
       (problem '(= (+ x 3) 8) '(= x 5) 2 t))
      (list
       (problem '(= (- x 10) 17) '(= x 27) 2)
       (problem '(= (+ x 91) 27) '(= x -64) 2)
       (problem '(= (+ x 5) 4) '(= x -1) 2)
       (problem '(= (/ x 3) 51) '(= x 153) 2)
       (problem '(= (* 7/2 x) 28) '(= x 8) 2))))
    (ch2-s6
     (values
      '() '(evaluate combine)
      (list
       (problem '(+ 3 (+ (* 2 x) 7)) '((+ (* 2 x) 10)) 2 t))
      (list
       (problem '(+ (+ 4 x) 3) '((+ x 7)) 2)
       (problem '(+ (+ 9 (* 7 x)) 3) '((+ 12 (* 7 x))) 2)
       (problem '(* -1/8 (* 56 x)) '((* -7 x)) 2)
       (problem '(/ (* -37 x) -1) '((* 37 x)) 2))))
    (ch2-s7
     (values
      '() '(evaluate invert)
      (list
       (problem '(= (+ (* 2 x) 3) 19) '(= x 8) 4 t)
       (problem '(= (- 10 x) 2) '(= x 8) 2 t))
      (list
       (problem '(= (+ (* 1/3 x) 5) 7) '(= x 6) 4)
       (problem '(= (- -4/3 x) -1/3) '(= x -1) 2))))
    (ch3-s1
     (values
      '() '(evaluate reorder combine)
      (list 
       (problem '(+ (* 5 (+ x 2)) 9) '((+ 19 (* 5 x))) 4 t))
      (list
       (problem '(+ (* 3 (+ x 5)) 11) '((+ 26 (* 3 x))) 4))))
    (ch3-s2
     (values
      '() '(evaluate reorder combine undo-minus)
      (list
       (problem '(- 9 (* 2 (- x 4))) '((+ (* -2 x) 17)) 5 t))
      (list
       (problem '(* 3 (+ x 7)) '((+ (* 3 x) 21)) 2)
       (problem '(- 11 (* 1 (+ (* 5 x) 3))) '((+ 8 (* -5 x))) 7))))
    (ch3-s4
     (values
      '() '(evaluate reorder combine undo-minus)
      (list
       (problem '(- (* 7 x) (* 5 x)) '((* 2 x)) 2 t)
       (problem '(+ (* 5 x) (- 6 (* 2 x))) '((+ (* 3 x) 6)) 2 t))
      (list
       (problem '(+ (* 5 x) (* 3 (- x 2))) '((- (* 8 x) 6)) 4))))
    (ch3-s5
     (values
      '() '(evaluate reorder combine canonicalize undo-minus)
      (list
       (problem '(+ (* 2 x) x) '((* 3 x)) 3 t)
       (problem '(- (+ (* 5 x) 3) (- 6 (* 2 x))) '((+ (* 7 x) -3)) 5 t))
      (list
       (problem '(- (* 6 x) (- x 1)) '((+ (* 5 x) 1)) 4)
       (problem '(- (* 8 (+ x 4)) (- x 5)) '((+ (* 7 x) 37)) 8))))
    (ch4-s1
     (values
      '() '(evaluate invert reorder combine canonicalize)
      (list
       (problem '(= (+ (* 2 x) (* 6 x)) 56) '(= x 7) 4 t))
      (list
       (problem '(= (- (* 7 x) (* 4 x)) 15) '(= x 5) 4)
       (problem '(= (- (* 10 x) x) 90) '(= x 10) 5))))
    (ch4-s2
     (values
      '() '(evaluate invert reorder combine canonicalize undo-minus)
      (list
       (problem '(= (+ (* 6 (- x 2)) (* 5 x)) 43) '(= x 5) 8 t))
      (list
       (problem '(= (+ (* 7 x) (* 3 x)) 40) '(= x 4) 4)
       (problem '(= 58 (+ (- 6 (* 14 x)) (* 12 x))) '(= x -26) 6)
       (problem '(= (+ (* 2 (- (* 3 x) 7)) (* 4 x)) 26) '(= x 4) 10))))
    (ch4-s3
     (values
      '() '(evaluate invert reorder combine canonicalize subtract undo-minus)
      (list
       (problem '(= (+ (* 3 x) 55) (* 8 x)) '(= x 11) 7 t))
      (list
       (problem '(= (+ (* 5 x) 27) (* 2 x)) '(= x -9) 7)
       (problem '(= (* -6 x) (- x 70)) '(= x 10) 9)
       (problem '(= (* 5 (- 9 x)) (* 4 (+ x 18))) '(= x -3) 14))))))

(defun short-tutor (&optional (subject (progn (princ "Subject number: ") (read)))
                    (condition (progn (format t "~@{~a~%~}"
                                              "1. Without All, Global Instruction"
                                              "2. Direct Demonstration, Global Instruction"
                                              "3. Without All, Without Global Instruction"
                                              "4. Direct Demonstration, Without Global Instruction") (read))))
  ;; setup the condition
  (setf *sticky-instructions* t)
  (ecase condition
    (1 (angela-experiment nil nil)
       (setf *no-solve-instructions* nil))
    (2 (angela-experiment t nil)
       (setf *no-solve-instructions* nil))
    (3 (angela-experiment nil nil)
       (setf *no-solve-instructions* t))
    (4 (angela-experiment t nil)
       (setf *no-solve-instructions* t)))
  ;; show calculator
  (unless (and calc::*calc-window*
               (window-shown-p calc::*calc-window*))
    (calc:calc))
  ;; start tutor
  (tutor subject
         (apply 'append
                (angela-short-intro 12 (not *no-solve-instructions*) (find condition '(2 4)))
                (build-session-sections 
                 '(ch1-s1 ch1-s2 ch1-s7 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch3-s5 ch4-s1 ch4-s2 ch4-s3)
                 'get-short-section))
         (list :angela-condition condition)))

(defun extract-problems-from-sets (sets)
  (loop for i in sets
       when (eq (first i) 'problem)
       append (problems (second i))))
