;;; replay a log file

(defpackage "REPLAY"
  (:export #:replay))
(in-package "REPLAY")

(defvar *log* nil
  "Holds the log")

(defvar *log-path* nil)

(defun slurp (s)
  "Return a list of sexpr's read from S"
  (do ((acc nil (nconc acc (list l)))
       (l (read s nil s) (read s nil s)))
      ((or (eq l 'experiment)
           (eq l s)) acc)))

(defun read-history (file)
  (let ((*package* (find-package "REPLAY")))
    (with-open-file (s file)
      ;; read the start line
      (slurp s))))

(defun click-calc-button (txt)
  (let ((buttons (tutor::all-subviews calc::*calc-window* 'button-dialog-item)))
    (ecase txt
      ((0 1 2 3 4 5 6 7 8 9 = + * -)
       (let ((b (find (prin1-to-string txt) buttons :key 'dialog-item-text :test 'string=)))
         (funcall (dialog-item-action-function b) b)))
      (:clear
       (let ((b (find "C" buttons :key 'dialog-item-text :test 'string=)))
         (funcall (dialog-item-action-function b) b)))
      (:divider
       (let ((b (find "/" buttons :key 'dialog-item-text :test 'string=)))
         (funcall (dialog-item-action-function b) b)))
      (calc::safe-/
       (let ((b (find "" buttons :key 'dialog-item-text :test 'string=)))
         (funcall (dialog-item-action-function b) b)))
      (:minus
       (let ((b (find "" buttons :key 'dialog-item-text :test 'string=)))
         (funcall (dialog-item-action-function b) b))))))

(defvar *annotate-clicks* nil
  "Set this to T to write the log to logfile.txt.annotated and
add click areas to the CLICK lines")

(defvar *click-view* nil
  "a view that shows where the subject clicked.")

(defun erase-click-view ()
  (when (and *click-view*
             (view-container *click-view*))
    (remove-subviews (view-container *click-view*) *click-view*)))

(defun show-click-location (x y window-space-p)
  (let ((p (if window-space-p
               (make-point x y)
               (convert-coordinates (make-point x y)
                                (tutor::tutor-eqn-view tutor::*tutor*)
                                tutor::*window*))))
    (erase-click-view)
    (setf *click-view* (tutor::draw-oval tutor::*window* 
                                         (- (point-h p) 10) (- (point-v p) 10)
                                         (+ (point-h p) 10) (+ (point-v p) 10)
                                         *red-color*))))

(defmacro with-annotated-file (stream &body body)
  `(with-open-file (,stream (merge-pathnames (format nil "~a.annotate" (file-namestring *log-path*))
                                             *log-path*)
                            :direction :output
                            :if-exists :append
                            :if-does-not-exist :create)
     ,@body))

(defgeneric log-write (type rest))

(defmethod log-write ((type t) rest)
  (with-annotated-file out
    (format out "~s~%" (cons type rest))))

(defmethod log-write ((type (eql 'click)) rest)
  (with-annotated-file out
    (format out "~s~%" (append (list type)
                               rest
                               (let ((p (convert-coordinates 
                                         (make-point (fourth rest) (fifth rest))
                                         (tutor::tutor-eqn-view tutor::*tutor*)
                                         tutor::*window*)))
                                 (tutor::find-object-at (point-h p) (point-v p)))))))

(defmethod log-write ((type (eql 'operation)) rest)
  (with-annotated-file out
    (format out "~s~%" (append (list type)
                               rest
                               (list (tutor::dump-equation (tutor::tutor-eqn-root tutor::*tutor*)))))))

(defvar *problem-before* nil
  "A list of problem states before the current one")
(defvar *problem-after* nil
  "A list of problem states after the current one")
(defvar *problem* nil
  "The problem tree")
(defvar *arg-ids* nil
  "the arg id alist used for the current submit phase.")

(defun find-box (target log-tree tutor-tree)
  (cond ((null log-tree)
         (error "can't find ~a - ~a ~a" target log-tree tutor-tree))
        ((atom log-tree)
         ;; skip
         nil)
        ((eq target (first log-tree))
         (first tutor-tree))
        ((and (listp log-tree)
              (= (length log-tree) 2))
         (find-box target (second log-tree) (second tutor-tree)))
        ((and (listp log-tree)
              (= (length log-tree) 4))
         (or (find-box target (third log-tree) (third tutor-tree))
             (find-box target (fourth log-tree) (fourth tutor-tree))))
        (t (error "uuugh."))))

(defun infix-problem (tree)
  (cond ((numberp tree)
         tree)
        ((symbolp tree)
         (intern (symbol-name tree) (find-package "REPLAY")))
        ((= (length tree) 2)
         (infix-problem (second tree)))
        (t (list (second tree)
                 (infix-problem (third tree))
                 (infix-problem (fourth tree))))))

(defun trees-match-p (log-tree)
  (let ((tutor-tree (infix-problem (tutor::box-to-tree (tutor::tutor-eqn-root tutor::*tutor*) t t)))
        (log-tree (infix-problem log-tree)))
    (equal log-tree tutor-tree)))

(defun no-duplicate-editable-types-p (root)
  (labels ((editable-type (box)
             (cond ((or (and (typep box 'tutor::answer-box)
                             (tutor::arg-editable (tutor::answer box)))
                        (tutor::arg1-editable-p box)
                        (tutor::op-editable-p box)
                        (tutor::arg2-editable-p box))
                    :arg)
                   ((tutor::box-editable-p box) :box)
                   ((tutor::op-arg-editable-p box) :op-arg)
                   ((tutor::arg-op-editable-p box) :arg-op)
                   (t (error "ugh")))))
    (let ((types (mapcar #'editable-type (tutor::editable-boxes root))))
      ;; a cheap hack to see if there are duplicates
      (values
       (= (length types) 
          (length (remove-duplicates types)))
       ;; are there non-arg duplicates?
       (= (length (remove :arg types))
          (length (remove-duplicates (remove :arg types))))))))

(defun associate-arg-boxes (fill-in assoc root)
  (let ((last (first (last fill-in)))
        (eqn (tutor::box-to-tree root t t))
        (tree (tutor::dump-equation (tutor::tutor-eqn-root tutor::*tutor*))))
    (when (eq (car last) 'submit)
      (loop for i in assoc
         collect (list (first i)
                       (let* ((box (find-box (second i) (fourth last) eqn))
                              (container (find box tree :key 'first))
                              ;; this is the arg id
                              (arg-id (loop for i in (second container)
                                           when (listp i)
                                           return (first i))))
                         arg-id))))))

(defun associate-args (fill-in root)
  "Return an alist of arg ids that appear in FILL-IN and arg ids that appear in ROOT."
  (let* (last
         (boxes (loop for i in fill-in
                   when (eq (car i) 'open-dialog)
                   do (setf last (fourth i))
                   when (and (eq (car i) 'input)
                             (or (find :op-editable (third (fifth i)))
                                 (find :arg1-editable (third (fifth i)))
                                 (find :arg2-editable (third (fifth i)))))
                   collect (list last (car (fifth i))))))
    (associate-arg-boxes fill-in boxes root)))

(defun rip-out-fill-in (log)
  "Return a part of the log up to submit. beware hitting back is
also a possibility which means we could be SOL."
  (loop for i in log
       collect i
       until (find (car i) '(back submit))))

(defun find-matching-editable (type log-id root duplicate-args-p)
  (format t "dups: ~a~%" duplicate-args-p)
  (let ((boxes (tutor::editable-boxes root)))
    (ecase type
      (arg
       (if duplicate-args-p
           (let ((id (second (find log-id *arg-ids* :key 'first))))
             (format t "HEY: ~a ~a ~a" *arg-ids* id log-id)
             id)
           (tutor::arg-id (first (tutor::editables (or (find-if (lambda (b)
                                                              (and (typep b 'tutor::answer-box)
                                                                   (tutor::arg-editable (tutor::answer b))))
                                                            boxes)
                                                   (find-if 'tutor::op-editable-p boxes)
                                                   (find-if 'tutor::arg1-editable-p boxes)
                                                   (find-if 'tutor::arg2-editable-p boxes)))))))
      (box 
       (tutor::box-id (find-if 'tutor::box-editable-p boxes)))
      (op-arg 
       (tutor::box-id (find-if 'tutor::op-arg-editable-p boxes)))
      (arg-op 
       (tutor::box-id (find-if 'tutor::arg-op-editable-p boxes))))))

(defstruct entry
  step-p fn)

(defvar *entries* (make-hash-table :test 'equal))

(defmacro define-entry ((name &optional step-p entry) &body body)
  (let ((ign (gensym)))
    `(setf (gethash ,(symbol-name name) *entries*)
           (make-entry
            :step-p ,step-p
            :fn ,(if entry
                     `(lambda (,entry) ,@body)
                     `(lambda (,ign) (declare (ignore ,ign)) ,@body))))))

(defvar *start-time* 0
  "The time the tutor was started")

(define-entry (:START-TUTOR nil entry) 
    (setf *start-time* (second entry)))
(define-entry (:TEXT-NEXT t)
    (tutor:execute-command :click :next))
(define-entry (:CALC t entry)
    (click-calc-button (third entry)))
(define-entry (:PROBLEM-START nil entry)
    (setf *problem* (third entry)
          *problem-after* nil
          *problem-before* nil))
(define-entry (:READ-INSTRUCTIONS-DONE t)
    (let ((b (or (find "Read the instructions, then click here." (tutor::all-subviews tutor::*window* 'button-dialog-item) :key 'dialog-item-text :test 'string=)
                 (find "Ready to go!" (tutor::all-subviews tutor::*window* 'button-dialog-item) :key 'dialog-item-text :test 'string=))))
      (when b
        (funcall (dialog-item-action-function b) b))))
(define-entry (:SELECT t entry)
    (tutor:execute-command :click (find-box (first (third entry)) 
                                            *problem*
                                            (tutor::box-to-tree (tutor::tutor-eqn-root tutor::*tutor*) t t))))
(define-entry (:UNSELECT t entry)
    (tutor:execute-command :click (find-box (first (third entry)) 
                                            *problem*
                                            (tutor::box-to-tree (tutor::tutor-eqn-root tutor::*tutor*) t t))))
(define-entry (:OPERATION t entry)
    (tutor:execute-command :click (intern (symbol-name (third entry)) "KEYWORD"))
    (when (tutor::tutor-submit-enabled tutor::*tutor*)
      (push *problem* *problem-before*)
      (setf *arg-ids* (associate-args (rip-out-fill-in *log*) (tutor::tutor-eqn-root tutor::*tutor*)))))

(define-entry (:OPEN-DIALOG t entry)
    ;;(assert (no-duplicate-editable-types-p (tutor::tutor-eqn-root tutor::*tutor*)))
    (multiple-value-bind (dup1 dup2) (no-duplicate-editable-types-p (tutor::tutor-eqn-root tutor::*tutor*))
    (tutor:execute-command :click (find-matching-editable (third entry) (fourth entry)
                                                          (tutor::tutor-eqn-root tutor::*tutor*)
                                                          (not (and dup1 dup2))))))
(define-entry (:INPUT-OPEN))
(define-entry (:INPUT-CLOSE))
(define-entry (:CLOSE-DIALOG))
;;     (tutor::clear-input-dialog))
(define-entry (:TYPED-INPUT t entry)
    (tutor:execute-command :type 
                           (case (third entry)
                             (backspace #\backspace)
                             (space #\space)
                             (t (third entry)))))
(define-entry (:INPUT-TYPE t entry)
    (ecase (third entry)
      (submit (tutor:execute-command :type #\Enter))))
(define-entry (:INPUT-CLICK t entry)
    (ecase (third entry)
      (submit (tutor:execute-command :type #\Enter))))
(define-entry (:INPUT))
(define-entry (:BOX-INPUT))
(define-entry (:SUBMIT nil entry)
    (setf *problem* (fourth entry)))
(define-entry (:BAD-SUBMIT))
(define-entry (:CLICK t entry)
    ;; FIXME: at some point that nil will be (sixth entry)
    (show-click-location (fourth entry) (fifth entry) nil))
(define-entry (:PROBLEM-END t)
    (tutor:execute-command :click :next-problem))
(define-entry (:KEYPRESS t))
(define-entry (:OPN-DURING-SUBMIT t))
(define-entry (:BREAK t)
    (tutor:execute-command :click :continue))
(define-entry (:BAD-FORWARD t))
(define-entry (:BAD-BACK t))
(define-entry (:INCORRECT-ANSWER t)
    (tutor:execute-command :click :next-problem))
(define-entry (:BACK t)
    (unless (tutor::tutor-submit-enabled tutor::*tutor*)
      (push *problem* *problem-after*))
    (setf *problem* (pop *problem-before*))
    (tutor:execute-command :click :back))
(define-entry (:FORWARD t)
    (push *problem* *problem-before*)
    (setf *problem* (pop *problem-after*))
    (tutor:execute-command :click :forward))
(define-entry (:LAST-CORRECT-STATE t)
    (tutor:execute-command :click :last-correct-state)
  (loop until (trees-match-p *problem*)
       do
       (push *problem* *problem-after*)
       (setf *problem* (pop *problem-before*))
       (assert *problem*)))

(define-entry (:USER-INSTRUCTION t entry)
    (unless (eq (third entry) :closed)
      (tutor::set-user-supplied-instruction-text (third entry))))
(define-entry (:WHATS-NEXT t)
    (tutor:execute-command :click :hint))
(define-entry (:ASK-EXPERIMENTER t)
    (tutor:execute-command :click :ask-experimenter))
(define-entry (:ASK-EXPERIMENTER-WAIT-DONE))
(define-entry (:ASK-EXPERIMENTER-RESUME t)
    (tutor:execute-command :click :ask-experimenter))
(define-entry (:LAST-CORRECT-STATE-FAIL t))

(defun do-log-entry (entry)
  (let ((e (gethash (symbol-name (car entry)) *entries*)))
    (if e
        (funcall (entry-fn e) entry)
        (error "unknown ~a" (car entry)))
    (entry-step-p e)))

(defun next-entry-step-p (log)
  (let ((e (gethash (symbol-name (caar log)) *entries*)))
    (and e (entry-step-p e))))

(defun log-file-type (log)
  "Given the log, determine what kind of log it is, either :REGULAR, :ANGELA, :INSTRUCTION-GENERATION."
  (cond ((find 'user-instruction log :key 'first)
         :instruction-generation)
        ((find 'read-instructions-done log :key 'first)
         :angela)
        (t :regular)))

(defun log-file-session (log)
  "Given the log, determine which session it logs."
  (let* ((line (find 'problem-start log :key 'first))
         (problem (infix-problem (third line))))
    (cond ((equal problem '(* (+ 5 7) 8)) 1)
          ((equal problem '(+ (* 5 (+ X 2)) 9)) 2)
          ((equal problem '(= (+ (* 2 X) (* 6 X)) 56)) 3)
          (t (error "unknown session")))))

(defun replay-window ()
  (find "Replay" (windows) :key 'window-title :test 'string=))

;; (defun add-text-to-window (s)
;;   (let ((edit (view-named 'edit (replay-window))))
;;     (set-dialog-item-text edit (concatenate 'string (dialog-item-text edit) (string #\Return) s))
;;     (ed-end-of-buffer edit)))

(defvar *play-process* nil)
(defvar *stop-playback* nil
  "Set to T to interrupt playback")

(defun stop-playback ()
  (when *play-process*
    (setf *stop-playback* t)
    (setf *play-process* nil)))

(defun my-sleep (miliseconds)
    (assert (= internal-time-units-per-second 1000))
  (let ((end (+ (get-internal-real-time) miliseconds)))
    (loop while (and (< (get-internal-real-time) end)
                     (not *stop-playback*)))))

(defun walk-log (&optional delay until)
  "Runs in a process and walks through the log at the pace the subject did it at when DELAY is nil."
  (labels ((now ()
             (second (first *log*))))
    (setf *stop-playback* nil)
    (let ((last (now))
          (until (if (listp until) 
                     until
                     (list until))))
      (loop while *log*
         until *stop-playback*
         until (and until 
                    (find (symbol-name (caar *log*)) (mapcar 'symbol-name until) :test 'string=))
         do (step-log)
           (my-sleep (or delay 
                         (- (now) last)))
           (setf last (now))))))

(defun play-log ()
  (unless *play-process*
    (setf *stop-playback* nil
          *play-process* (process-run-function "replay" 'walk-log))))

(defun ff-log ()
  (unless *play-process*
    (setf  *stop-playback* nil
           *play-process* (process-run-function "replay" 'walk-log 500))))

(defun in-secs (n)
  (/ (- n *start-time*) 1000))

(defun step-log ()
  (labels ((step ()
             (let ((i (pop *log*)))
               (let ((*package* (find-package "REPLAY")))
                 (format t "(~s ~,1F~{ ~s~})~%" (first i) (in-secs (second i)) (cddr i)))
               (prog1
                (do-log-entry i)
                (let ((*package* (find-package "REPLAY")))
                  (when *annotate-clicks*
                    (log-write (first i) (rest i))))))))
    (erase-click-view)
    ;; go forward til we hit an action
    (loop until (step))
    ;; skip over any non-actions after that
    (loop while (and *log*
                     (not (next-entry-step-p *log*)))
       do (step))))

(defun next-problem ()
  (when (string= (symbol-name (caar *log*)) "PROBLEM-START")
    (step-log))
  (walk-log 0 :problem-start))

(defun next-section ()
  (when (find (symbol-name (caar *log*)) '("BREAK" "START-TUTOR") :test 'string=)
    (step-log))
  (walk-log 0 '(:break :start-tutor)))

(defun create-replay-window ()
  (let ((panel (make-instance 'window
                              :window-title "Replay"
                              :window-type :document
                              :view-font '("Monaco" 14)
                              :view-size (make-point 113 240))))
;;                               :view-subviews
;;                               (list
;;                                (make-instance 'button-dialog-item
;;                                               :view-size #@(100 20)
;;                                               :view-position #@(150 2)
;;                                               :dialog-item-text "Step"
;;                                               :dialog-item-action (tutor::button-fn (step-log)))
;;                                (make-instance 'editable-text-dialog-item
;;                                               :view-nick-name 'edit
;;                                               :view-size #@(398 170)
;;                                               :view-position #@(1 30))))))
    (tutor::create-layout `(:vbox :contents ((:button :text "Step" :action ,(tutor::button-fn (step-log)))
                                             (:button :text "Play" :action ,(tutor::button-fn (play-log)))
                                             (:button :text "Stop" :action ,(tutor::button-fn (stop-playback)))
                                             (:button :text "FF" :action ,(tutor::button-fn (play-log)))
                                             (:button :text "Next Problem" :action ,(tutor::button-fn (next-problem)))
                                             (:button :text "Next Section" :action ,(tutor::button-fn (next-section)))))
                          panel)))
;;     panel))
    

(defun replay (&optional (file (choose-file-dialog)) type-subject)
  (format t "Reading log file ~s ..." file)
  (setf *log* (read-history file)
        *log-path* file)
  (format t "Done.~%")
  (let* ((type (log-file-type *log*))
         (session (log-file-session *log*))
         (fn (case session 
               (1 'tutor::angela-session1)
               (2 'tutor::angela-session2)
               (3 'tutor::angela-session3))))
    (ecase type
      (:instruction-generation (apply fn t type-subject))
      (:angela (apply fn nil type-subject))
      (:regular 
       (error "This log file is not yet supported")))
    (unless (replay-window)
      (create-replay-window))))

(defun annotate-all ()
  (labels ((replay-it-all (base id)
             (loop for i from 1 to 3 do
                  (replay (merge-pathnames base (format nil "history-~d-session~d.txt" id i)) '(1 0))
                  (ignore-errors
                    (walk-log 0 :never-reached))
                  (window-close tutor::*window*))))
    (let ((base1 #p"Macintosh HD:Users:sabetts:tutor-forms:angela-exp:archive:logs_version1:")
          (base2 #p"Macintosh HD:Users:sabetts:tutor-forms:angela-exp:archive:logs_version2:sessions:"))
      (setf *annotate-clicks* t)
      ;; 1 session 2 fails, 2 session 2 fails, 3 session 2 unknown session, 17 unknown session
;;       (loop for i in '( 18 192 202) do
;;            (replay-it-all base1 i))
      ;; 25 unknown sess
      (loop for i in '( 40 41 42 44 45 352) do
           (replay-it-all base2 i)))))

(defun check-all ()
  (labels ((check-it (base id)
             (loop for i from 1 to 3 do
                  (let* ((f1 (merge-pathnames base (format nil "history-~d-session~d.txt" id i)))
                         (h1 (and (probe-file f1) (read-history f1)))
                         (l1 (or (and h1 (length h1)) 0))
                         (f2 (merge-pathnames base (format nil "history-~d-session~d.txt.annotate" id i)))
                         (h2 (and (probe-file f2) (read-history f2)))
                         (l2 (or (and h2 (length h2)) 0)))
                    (when (= l1 l2)
                      (format t "length mismatch: history-~d-session~d.txt. Compensating. Mark 10-4, roger. ~%" id i))
                    ;; dump the annotated part plus the unannotated part into a new file
                    (with-open-file (stream (merge-pathnames base (format nil "history-~d-session~d.final" id i))
                                            :direction :output
                                            :if-exists :append
                                            :if-does-not-exist :create)
                      (format stream "~{~s~%~}" h2)
                      (format stream ";; Annotation Ends Here~%")
                      (format stream "~{~s~%~}" (subseq h1 l2)))))))
    (let ((base1 #p"Macintosh HD:Users:sabetts:tutor-forms:angela-exp:archive:logs_version1:")
          (base2 #p"Macintosh HD:Users:sabetts:tutor-forms:angela-exp:archive:logs_version2:sessions:"))
      ;; 1 session 2 fails, 2 session 2 fails, 3 session 2 unknown session
      (loop for i in '(1 2 3 4 5 7 8 9 10 11 12 13 14 15 16 17 18 192 202) do
           (check-it base1 i))
      (loop for i in '(25 26 27 28 29 30 31 32 33 34 352 36 37 38 39 40 41 42 44 45 352) do
           (check-it base2 i)))))
