;;; Annotate sections of a wave file with text. Dump the text in a timestamped format.
(require 'quickdraw)

(defpackage :annotate-wav
  (:export #:do-child))

(in-package :annotate-wav)

(defvar *playback-program* "./PlayBufferedSoundFile")
(defvar *playback-file* nil)
(defvar *cycles-per-sec*)-

(defun macify-path (p)
  (concatenate 'string "Macintosh HD"
               (substitute #\: #\/ p)))

(defun unixify-path (p)
  (let* ((s (namestring p))
         (pos (+ 12 (search "Macintosh HD" s))))
    (nsubstitute #\/ #\: (subseq s pos))))

(defclass wav-window (window)
  ())

(defclass wav-editable (editable-text-dialog-item)
  ())

(defstruct tag
  start end text)

(defclass wav-view (view)
  ((sel-start :initform nil :accessor sel-start)
   (sel-end :initform nil :accessor sel-end)
   (data :initarg :data :initform nil :accessor wav-data)
   (scroll :initform 0 :accessor wav-scroll)
   (tags :initarg :tags :initform nil :accessor wav-tags)))

(defmethod window-size-parts ((w wav-window))
  (let ((wav (view-named 'wave w))
        (scroll (view-named 'scroll w)))
    (when (and wav scroll)
      (set-scroll-bar-length scroll (point-h (view-size w)))
      (setf (scroll-bar-page-size scroll) (truncate (point-h (view-size w)) 2))
      (set-view-size wav (point-h (view-size w)) (point-v (view-size wav))))))

(defun wav-select-tag (wav edit tag)
  (setf (sel-start wav) (tag-start tag)
        (sel-end wav) (tag-end tag))
  (set-dialog-item-text edit (tag-text tag))
  (setf (wav-tags wav) (remove tag (wav-tags wav))))

(defun next-tag (point tags)
  "Return the next tag after point."
  (let ((tmp (sort (copy-list tags) (lambda (a b) (< (tag-start a) (tag-start b))))))
    (find point tmp :key 'tag-start :test '<=)))

(defun previous-tag (point tags)
  "Return the next tag after point."
  (let ((tmp (sort (copy-list tags) (lambda (a b) (> (tag-start a) (tag-start b))))))
    (find point tmp :key 'tag-start :test '>=)))

(defun skip-to-tag (wav dir)
  (let* ((start (sel-start wav))
         (end (sel-end wav))
         (next (funcall (if (eq dir :next) 'next-tag 'previous-tag) (or start 0) (wav-tags wav)))
         (scroll (view-named 'scroll (view-container wav))))
    (when (and start end)
      (maybe-add-tag wav (min start end) (max start end)))
    (when next
      (set-scroll-bar-setting scroll (max 0 (- (tag-start next) (truncate (point-h (view-size wav)) 2))))
      (update-wav-view scroll)
      (wav-select-tag wav (view-named 'edit (view-container wav)) next))))

(defun skip-to-next-tag (wav)
  (skip-to-tag wav :next))

(defun skip-to-previous-tag (wav)
  (skip-to-tag wav :previous))

(defmethod view-click-event-handler ((view wav-view) where)
  ;;   (format t "click! ~d ~a~%"(point-h where) (wav-tags view))
  (let* ((loc (+ (wav-scroll view) (point-h where)))
         (tag (find-tag view loc)))
    (if (sel-start view)
        (unless tag
          (setf (sel-end view) loc))
        (if tag
            (wav-select-tag view (view-named 'edit (view-container view)) tag)
            (setf (sel-start view) loc))))
  (invalidate-view view))

;; (defmethod view-key-event-handler ((w wav-window) ch)
;;   (view-key-event-handler (view-named 'wave w) ch))

(defun maybe-add-tag (wav start end)
  (let ((e (view-named 'edit (view-container wav))))
    (when (> (length (dialog-item-text e)) 0)
      (push (make-tag :text (dialog-item-text e)
                      :start start
                      :end end)
            (wav-tags wav))
      (set-dialog-item-text e "")
      ;; save the tags file everytime a change is made. Seems like a
      ;; good thing...
      (save-tags wav)
      )))
                    
(defrecord Event
    (what integer)
  (message longint)
  (when longint)
  (where point)
  (modifiers integer))

(defvar *region-offset* 0
  "current region offset for playing subregions.")
(defparameter +region-length+ 5)
(defparameter +region-inc+ 4)

(defun play-current-subregion (view)
  (when (and (sel-start view) (sel-end view))
    (let ((start (+ (* *region-offset* *cycles-per-sec*)
                    (min (sel-start view) (sel-end view))))
          (end (max (sel-start view) (sel-end view))))
      (playback-play-region start (min (+ start (* +region-length+ *cycles-per-sec*)) end))
      (incf *region-offset* +region-inc+)
      (when (>= (+ start (* +region-length+ *cycles-per-sec*)) end)
        (setf *region-offset* 0)))))

(defun play-current-region (view)
  ;; reset the offset
  (setf *region-offset* 0)
  (when (and (sel-start view) (sel-end view))
    (playback-play-region (min (sel-start view) (sel-end view)) (max (sel-start view) (sel-end view)))))

(defmethod view-key-event-handler :around ((editable wav-editable) key)
  (declare (ignore key))
  (let* ((view (view-named 'wave (view-container editable)))
         (edit (view-named 'edit (view-container editable)))
         (keystroke (keystroke-name (event-keystroke (rref *current-event* :event.message) (rref *current-event* :event.modifiers))))
         (ch (if (listp keystroke)
                 (car (last keystroke))
                 keystroke)))
    (cond ((find ch '(#\Escape #\Enter #\Return) :test 'char=)
           (when (and (sel-start view) (sel-end view))
             (maybe-add-tag view (min (sel-start view) (sel-end view)) (max (sel-start view) (sel-end view))))
           (setf (sel-start view) nil
                 (sel-end view) nil)
           (invalidate-view view))
          ((char= ch #\Tab)
           (if (and (listp keystroke)
                    (find :shift (butlast keystroke)))
               (skip-to-previous-tag view)
               (skip-to-next-tag view))
           (when (string= (dialog-item-text edit) "")
             (play-current-region view))
           (invalidate-view view))
          ((and (char= ch #\d)
                (listp keystroke)
                (find :meta (butlast keystroke)))
           (dump-tags (wav-tags view) (choose-new-file-dialog)))
          ((and (char= ch #\r)
                (listp keystroke)
                (find :control (butlast keystroke)))
           (play-current-region view))
          ((and (char= ch #\r)
                (listp keystroke)
                (find :meta (butlast keystroke)))
           (play-current-subregion view))
          (t (call-next-method)))))

(defun data-length (data)
  (apply '+ (map 'list 'length data)))

(defconstant +data-len+ 10000000)
(defconstant +max-data-scroll-length+ 80000
  "shrink the wave file fit within this many data points.")

(defun data-aref (data nth)
  (let ((a (mod nth +data-len+))
        (b (truncate nth +data-len+)))
    (abs (- (aref (aref data b) a) 127))))

(defun find-tag (wav x)
  (find-if (lambda (tag)
             (and (>= x (tag-start tag))
                  (<= x (tag-end tag))))
             (wav-tags wav)))

(defun draw-tags (wav ofs)
    (dolist (i (wav-tags wav))
      (with-fore-color (if (zerop (length (tag-text i))) *red-color* *green-color*)
      (frame-rect wav (- (tag-start i) ofs) 5 (- (tag-end i) ofs) (- (point-v (view-size wav)) 5)))))

(defmethod view-draw-contents ((view wav-view))
  (let* ((ofs (wav-scroll view))
         (len (point-h (view-size view)))
         (h (point-v (view-size view)))
         (paint-len (max (- (data-length (wav-data view)) ofs) 0)))
    (loop for i from ofs below (+ ofs (min len paint-len))
       for j from 0 below (min len paint-len)
       ;; erase any previous line
       do (with-fore-color *white-color*
            (move-to view j 0)
            (line-to view j h))
       do (move-to view j (- (truncate h 2) (data-aref (wav-data view) i)))
       do (line-to view j (+ (truncate h 2) (data-aref (wav-data view) i))))
    ;; clear the remainder
    (when (< paint-len len)
      (with-fore-color *white-color*
        (paint-rect view paint-len 0 len h)))
    ;; draw the selection
    (when (sel-start view)
      (let ((end (or (sel-end view) (1+ (sel-start view)))))
        (invert-rect view (- (min (sel-start view) end) ofs) 0 (- (max (sel-start view) end) ofs) h)))
    (draw-tags view ofs)
    ))

(defun update-wav-view (scroll-bar)
  (let ((scrollee (scroll-bar-scrollee scroll-bar))
        (setting (scroll-bar-setting scroll-bar)))
    (setf (wav-scroll scrollee) setting)
    (invalidate-view scrollee)))

;; (defun make-data ()
;;   (let ((a (make-array 400)))
;;     (loop for i from 0 below 400
;;           do (setf (aref a i) i))
;;     a))

(defun read-raw-data (file)
  "Return an array of bytes."
    (with-open-file (s file :element-type '(unsigned-byte 8))
      (let ((top (make-array (1+ (truncate (file-length s) +data-len+)) :fill-pointer 0)))
        (loop while (< (file-position s) (file-length s))
           do (format t "okay - ~d ~d ~d~%" (file-position s) (file-length s) (length top))
           do (let ((data (make-array (min +data-len+ (1+ (- (file-length s) (file-position s)))) :element-type '(unsigned-byte 8) :initial-element 0)))
                (read-sequence data s)
                (vector-push-extend data top)))
      top)))

(defun data-subseq (data start end &optional (type 'vector))
  (let ((start-idx (truncate start +data-len+))
        (start-ofs (mod start +data-len+))
        (end-idx (truncate end +data-len+))
        (end-ofs (mod end +data-len+)))
    (if (= start-idx end-idx)
        (coerce (subseq (aref data start-idx) start-ofs end-ofs) type)
        (concatenate type
                     (subseq (aref data start-idx) start-ofs +data-len+)
                     (subseq (aref data end-idx) 0 end-ofs)))))


(defun fit-data-with-limit (data)
  (if (<= (data-length data) +max-data-scroll-length+)
      data
      (make-array 1 :initial-element (coerce (let ((inc (/ (data-length data) +max-data-scroll-length+)))
                                               (loop for i from 0 below (data-length data) by inc
                                                  collect (let* ((seq (data-subseq data (floor i) (floor (+ i inc)) 'list))
                                                                 (max (apply 'max seq))
                                                                 (min (apply 'min seq)))
                                                            (if (> (abs (- 128 max)) (abs (- 128 min)))
                                                                max min)))) 
                                             'vector))))

(defun annotate (data retag)
  (let* ((w (make-instance 'wav-window
                           :view-size (make-point 1000 400)))
         (v (make-instance 'wav-view
                           :view-nick-name 'wave
                           :view-container w
                           :view-position #@(0 0)
                           :view-size (make-point 1000 256)
                           :data data
                           ;; resume work if we can, otherwise generate the tags.
                           :tags (or (and (not retag) (maybe-load-tags *playback-file*))
                                     (generate-blank-tags data))))
         (s (make-instance 'scroll-bar-dialog-item
                           :view-nick-name 'scroll
                           :direction :horizontal
                           :view-container w
                           :view-position #@(0 256)
                           :length 1000
                           :min 0
                           :max (data-length data)
                           :page-size 500
                           :scroll-size 20
                           :scrollee v
                           :dialog-item-action 'update-wav-view))
         (e (make-instance 'wav-editable
                           :allow-tabs t
                           :word-wrap-p t
                           :view-nick-name 'edit
                           :view-container w
                           :view-position #@(10 280)
                           :view-size #@(500 100)))
         (help (make-instance 'static-text-dialog-item
                              :view-container w
                              :view-position #@(520 280)
                              :view-size #@(500 100)
                              :view-nick-name 'help
                              :dialog-item-text (format nil "Escape - clear selection~%Tab - next tag~%Shift+Tab - previous tag~%Control+r - play selection~%Option+d - dump tags")))
         (but (make-instance 'button-dialog-item
                             :view-container w
                             :view-position #@(520 370)
                             :dialog-item-text "Recreate Tags"
                             :dialog-item-action 'recreate-tags)))
    (values w but help e s)))

(defun save-tags (wav)
  (with-open-file (s (merge-pathnames (make-pathname :type "tags") (macify-path *playback-file*)) :direction :output :if-exists :overwrite)
    (format s "~s~%" (wav-tags wav))))

(defun maybe-load-tags (file)
  (handler-case
      (with-open-file (s (merge-pathnames (make-pathname :type "tags") (macify-path file)) :direction :input)
        (write-line "Loaded tags.")
        (let ((*package* (find-package :annotate-wav)))
          (read s)))
    (file-error (c) (declare (ignore c))
                (write-line "Starting with a blank tags file.")
                nil)))

;; sox child3.wav -t .raw -u -b -c 1 -r 1000 raw.raw
(defun do-child (&optional (filename (choose-file-dialog)) retag)
  (playback-set-file (unixify-path filename))
  (let* ((d1 (read-raw-data (merge-pathnames (make-pathname :type "raw") filename)))
         (d2 (fit-data-with-limit d1)))
    (setf *cycles-per-sec* (* 1000 (/ +max-data-scroll-length+ (data-length d1))))
    (annotate d2 retag)))

(defmacro with-playback ((stream) &body body)
  `(with-open-file (,stream #P"Macintosh HD:tmp:mclcmd" :direction :output :if-exists :overwrite)
     ,@body))

(defun playback-set-file (file)
  (setf *playback-file* file))

(defun playback-play-region (start end)
  (let ((s1 (/ start *cycles-per-sec*))
        (s2 (/ end *cycles-per-sec*)))
    (with-playback (s)
      (format s "~a ~a ~,1f ~,1f~%" *playback-program* *playback-file* s1 (+ s2 1)))))

;;; Auto generate the speech tags

(defvar +sig-height+ 5)
(setf (documentation '+sig-height+ 'variable) "activation height. How loud does a sound bite have to be to start a speech block?")
(defvar +cut-off+ 10)
(setf (documentation '+cut-off+ 'variable) "after this much inactivity, end the speech block")
(defvar +min-length+ 5)
(setf (documentation '+min-length+ 'variable) "An audio blip must last at least this long for it to be considered a speech block.")
(defvar +min-height+ 10)
(setf (documentation '+min-height+ 'variable) "Alternatively, if an audio blip has a spike at least this high then consider it a speech block no matter how long it is.")
(defvar +pad-length+ 5)
(setf (documentation '+pad-length+ 'variable) "Pad the speech block at both ends by this much.")

(defun recreate-tags (button)
  (let ((wav (view-named 'wave (view-container button))))
    (y-or-n-dialog "This will clear all existing tags. continue?" :no-text nil)
    (macrolet ((setvar (symbol)
                 `(setf ,symbol (read-from-string (get-string-from-user (documentation ',symbol 'variable)
                                                                        :initial-string (prin1-to-string ,symbol))))))
      (setvar +sig-height+)
      (setvar +cut-off+)
      (setvar +min-length+)
      (setvar +min-height+)
      (setvar +pad-length+)
      (setf (wav-tags wav) (generate-blank-tags (wav-data wav)))
      (invalidate-view wav))))

(defun generate-blank-tags (data)
  (let (acc)
    (loop for i from 0 below (data-length data)
       with state = 'start
       with last  = 0
       with start = 0
       with max = 0
       do (ecase state
            (start (when (>= (abs (data-aref data i)) +sig-height+)
                     (setf state 'end
                           last i
                           start i)))
            (end (when (>= (abs (data-aref data i)) +sig-height+)
                   (setf last i
                         max (max max (abs (data-aref data i)))))
                 (when (> (- i last) +cut-off+)
                   (when (or (>= (- last start) +min-length+)
                             (> max +min-height+))
                     (push (make-tag :start (max 0 (- start +pad-length+)) :end (+ last +pad-length+) :text "") acc))
                   (setf state 'start)))))
    acc))

(defun dump-tags (tags file)
  (with-open-file (s file :direction :output :if-exists :overwrite)
    (loop for i in (sort (copy-list tags) (lambda (a b) (< (tag-start a) (tag-start b))))
       do (format s "~,1f	~a~%" (/ (tag-start i) *cycles-per-sec*) (tag-text i))
       do (format s "~,1f	 ~%" (/ (tag-end i) *cycles-per-sec*)))))
